"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import pytest

from cfnlint.rules.functions.SubNotJoin import SubNotJoin


@pytest.fixture(scope="module")
def rule():
    rule = SubNotJoin()
    yield rule


@pytest.fixture
def template():
    return {
        "Resources": {
            "MyResource": {
                "Type": "AWS::S3::Bucket",
            },
            "CDK": {
                "Type": "AWS::CDK::Metadata",
            },
        },
    }


@pytest.mark.parametrize(
    "name,instance,schema,expected",
    [
        (
            "Invalid Fn::Join with an empty string",
            {"Fn::Join": ["", ["foo", "bar"]]},
            {"type": "string"},
            [],
        ),
    ],
)
def test_validate(name, instance, schema, expected, rule, validator):
    errs = list(rule.validate(validator, schema, instance, {}))
    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.route53.RecordSetName import RecordSetName


@pytest.fixture(scope="module")
def rule():
    rule = RecordSetName()
    yield rule


@pytest.mark.parametrize(
    "instance,expected",
    [
        (
            {
                "HostedZoneName": "bar.",
                "Name": "foo.bar",
            },
            [],
        ),
        (
            {
                "HostedZoneName": "bar.",
                "Name": "bar",
            },
            [],
        ),
        (
            {
                "HostedZoneName": "bar.",
                "Name": {"Ref": "pName"},
            },
            [],
        ),
        (
            {
                "HostedZoneName": {"Ref": "pHostedZoneName"},
                "Name": "foo.bar",
            },
            [],
        ),
        (
            {
                "HostedZoneName": "bar",
                "Name": "foo.bar.",
            },
            [
                ValidationError(
                    "'bar' must end in a dot",
                    path=deque(["HostedZoneName"]),
                )
            ],
        ),
        (
            {
                "HostedZoneName": "bar.",
                "Name": "bar.foo.",
            },
            [
                ValidationError(
                    "'bar.foo.' must be a subdomain of or equal to 'bar.'",
                    path=deque(["Name"]),
                )
            ],
        ),
        (
            {
                "HostedZoneName": "bar.",
                "Name": "foobar.",
            },
            [
                ValidationError(
                    "'foobar.' must be a subdomain of or equal to 'bar.'",
                    path=deque(["Name"]),
                )
            ],
        ),
    ],
)
def test_validate(instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))

    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.s3.AccessControlObsolete import AccessControlObsolete


@pytest.fixture(scope="module")
def rule():
    rule = AccessControlObsolete()
    yield rule


@pytest.mark.parametrize(
    "instance,expected",
    [
        (
            {
                "AccessControl": "AuthenticatedRead",
            },
            [
                ValidationError(
                    (
                        "'AccessControl' is a legacy property. "
                        "Consider using 'AWS::S3::BucketPolicy' instead"
                    ),
                    path=deque(["AccessControl"]),
                )
            ],
        ),
        (
            {
                "AccessControl": {"Ref": "AWS::NoValue"},
            },
            [],
        ),
        (
            {},
            [],
        ),
    ],
)
def test_validate(instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))

    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from typing import Any

import cfnlint.data.schemas.extensions.aws_route53_healthcheck
from cfnlint.jsonschema import ValidationError
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails


class HealthCheckHealthCheckConfigTypeInclusive(CfnLintJsonSchema):
    id = "E3661"
    shortdesc = (
        "Validate Route53 health check has AlarmIdentifier when using CloudWatch"
    )
    description = (
        "When 'Type' is 'CLOUDWATCH_METRIC' you must specify 'AlarmIdentifier'"
    )
    tags = ["resources"]

    def __init__(self) -> None:
        super().__init__(
            keywords=[
                "Resources/AWS::Route53::HealthCheck/Properties/HealthCheckConfig"
            ],
            schema_details=SchemaDetails(
                module=cfnlint.data.schemas.extensions.aws_route53_healthcheck,
                filename="healthcheckconfig_type_inclusive.json",
            ),
        )

    def message(self, instance: Any, err: ValidationError) -> str:
        return err.message



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

import json
import unittest
from collections import deque
from typing import List

import jsonpatch

from cfnlint.decode import decode_str
from cfnlint.helpers import FUNCTIONS
from cfnlint.jsonschema import CfnTemplateValidator, ValidationError
from cfnlint.template import Template


def build_template(patch: jsonpatch.JsonPatch | None) -> Template:
    """
    Build a template object from a dict
    """
    template_obj = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Conditions": {
            "IsUsEast1": {"Fn::Equals": ["true", {"Ref": "AWS::Region"}]},
            "IsNotUsEast1": {"Fn::Not": [{"Condition": "IsUsEast1"}]},
        },
        "Resources": {
            "MyResource": {
                "Type": "AWS::Test::Type",
                "Properties": {
                    "Name": "Name",
                    "Tags": [
                        {"Key": "Name", "Value": "Value"},
                    ],
                },
            }
        },
    }

    if patch is not None:
        jsonpatch.apply_patch(template_obj, patch, True)

    template, _ = decode_str(json.dumps(template_obj))
    if template:
        return Template("", template, regions=["us-east-1"])

    raise Exception("Template is invalid")


def build_schema(patch: jsonpatch.JsonPatch | None):
    schema = {
        "properties": {
            "Name": {
                "type": "string",
            },
            "Tags": {
                "type": "array",
                "items": {
                    "type": "object",
                    "properties": {
                        "Key": {"type": "string"},
                        "Value": {"type": "string"},
                    },
                },
            },
        },
        "type": "object",
    }

    if patch is not None:
        jsonpatch.apply_patch(schema, patch, True)

    return schema


class TestValidatorCfnConditions(unittest.TestCase):
    """
    Integration testing of the CfnTemplateValidator class
    with a Template that allows us to exhaust scenario
    """

    def run_test(
        self,
        template_patches: jsonpatch.JsonPatch,
        schema_patches: jsonpatch.JsonPatch,
        expected_errs: List[ValidationError],
    ):
        """
        Run the test
        """
        cfn = build_template(template_patches)
        schema = build_schema(schema_patches)

        props = (
            cfn.template.get("Resources", {})
            .get("MyResource", {})
            .get("Properties", {})
        )

        context = cfn.context.evolve(functions=FUNCTIONS)
        validator = CfnTemplateValidator(schema=schema, cfn=cfn, context=context)
        errs = list(validator.iter_errors(props))

        self.assertEqual(len(errs), len(expected_errs), f"Received: {errs!r}")
        for i, err in enumerate(errs):
            self.assertEqual(expected_errs[i].message, err.message)
            self.assertEqual(expected_errs[i].path, err.path)
            self.assertEqual(expected_errs[i].validator, err.validator)
            self.assertEqual(expected_errs[i].validator_value, err.validator_value)

    def test_required(self):
        schema_patch = jsonpatch.JsonPatch(
            patch=[{"op": "add", "path": "/required", "value": ["Name"]}],
        )
        self.run_test(
            template_patches=None,
            schema_patches=schema_patch,
            expected_errs=[],
        )

        self.run_test(
            template_patches=jsonpatch.JsonPatch(
                patch=[
                    {"op": "remove", "path": "/Resources/MyResource/Properties/Name"}
                ],
            ),
            schema_patches=schema_patch,
            expected_errs=[
                ValidationError(
                    "'Name' is a required property",
                    validator="required",
                    validator_value=["Name"],
                ),
            ],
        )

        self.run_test(
            template_patches=jsonpatch.JsonPatch(
                patch=[
                    {
                        "op": "replace",
                        "path": "/Resources/MyResource/Properties/Name",
                        "value": {
                            "Fn::If": [
                                "IsUsEast1",
                                "Name",
                                {"Ref": "AWS::NoValue"},
                            ]
                        },
                    }
                ],
            ),
            schema_patches=schema_patch,
            expected_errs=[
                ValidationError(
                    "'Name' is a required property",
                    validator="required",
                    validator_value=["Name"],
                ),
            ],
        )

    def test_unique_items(self):
        schema_patch = jsonpatch.JsonPatch(
            [
                {"op": "add", "path": "/properties/Tags/uniqueItems", "value": True},
            ]
        )
        self.run_test(
            template_patches=None,
            schema_patches=schema_patch,
            expected_errs=[],
        )

        self.run_test(
            template_patches=jsonpatch.JsonPatch(
                [
                    {
                        "op": "add",
                        "path": "/Resources/MyResource/Properties/Tags/-",
                        "value": {"Key": "Name", "Value": "Value"},
                    },
                ]
            ),
            schema_patches=schema_patch,
            expected_errs=[
                ValidationError(
                    message=("array items are not unique"),
                    path=deque(["Tags"]),
                    validator="uniqueItems",
                    validator_value=True,
                ),
            ],
        )

        self.run_test(
            template_patches=jsonpatch.JsonPatch(
                [
                    {
                        "op": "add",
                        "path": "/Resources/MyResource/Properties/Tags/-",
                        "value": {
                            "Fn::If": [
                                "IsUsEast1",
                                {"Key": "Name", "Value": "Value"},
                                {"Key": "Foo", "Value": "Bar"},
                            ]
                        },
                    },
                ]
            ),
            schema_patches=schema_patch,
            expected_errs=[
                ValidationError(
                    message=("array items are not unique"),
                    path=deque(["Tags"]),
                    validator="uniqueItems",
                    validator_value=True,
                ),
            ],
        )

        self.run_test(
            template_patches=jsonpatch.JsonPatch(
                [
                    {
                        "op": "replace",
                        "path": "/Resources/MyResource/Properties/Tags",
                        "value": [
                            {
                                "Fn::If": [
                                    "IsUsEast1",
                                    {"Key": "Name", "Value": "Value"},
                                    {"Ref": "AWS::NoValue"},
                                ],
                            },
                            {
                                "Fn::If": [
                                    "IsUsEast1",
                                    {"Ref": "AWS::NoValue"},
                                    {"Key": "Name", "Value": "Value"},
                                ],
                            },
                        ],
                    },
                ]
            ),
            schema_patches=schema_patch,
            expected_errs=[],
        )

        self.run_test(
            template_patches=jsonpatch.JsonPatch(
                [
                    {
                        "op": "replace",
                        "path": "/Resources/MyResource/Properties/Tags",
                        "value": [
                            {
                                "Fn::If": [
                                    "IsUsEast1",
                                    {"Key": "Name", "Value": "Value"},
                                    {"Ref": "AWS::NoValue"},
                                ],
                            },
                            {
                                "Fn::If": [
                                    "IsNotUsEast1",
                                    {"Ref": "AWS::NoValue"},
                                    {"Key": "Name", "Value": "Value"},
                                ],
                            },
                        ],
                    },
                ]
            ),
            schema_patches=schema_patch,
            expected_errs=[
                ValidationError(
                    message=("array items are not unique"),
                    path=deque(["Tags"]),
                    validator="uniqueItems",
                    validator_value=True,
                ),
            ],
        )

    def test_dependencies(self):
        schema_patch = jsonpatch.JsonPatch(
            [
                {"op": "add", "path": "/dependencies", "value": {"Name": ["Tags"]}},
            ]
        )
        self.run_test(
            template_patches=None,
            schema_patches=schema_patch,
            expected_errs=[],
        )

        self.run_test(
            template_patches=jsonpatch.JsonPatch(
                [
                    {"op": "remove", "path": "/Resources/MyResource/Properties/Tags"},
                ]
            ),
            schema_patches=schema_patch,
            expected_errs=[
                ValidationError(
                    message="'Tags' is a dependency of 'Name'",
                    path=deque([]),
                    validator="dependencies",
                    validator_value={"Name": ["Tags"]},
                ),
            ],
        )

        self.run_test(
            template_patches=jsonpatch.JsonPatch(
                [
                    {
                        "op": "replace",
                        "path": "/Resources/MyResource/Properties/Tags",
                        "value": {
                            "Fn::If": [
                                "IsUsEast1",
                                {"Ref": "AWS::NoValue"},
                                [
                                    {
                                        "Fn::If": [
                                            "IsUsEast1",
                                            {"Ref": "AWS::NoValue"},
                                            {"Key": "Name", "Value": "Value"},
                                        ]
                                    },
                                ],
                            ]
                        },
                    },
                ]
            ),
            schema_patches=schema_patch,
            expected_errs=[
                ValidationError(
                    message="'Tags' is a dependency of 'Name'",
                    path=deque([]),
                    validator="dependencies",
                    validator_value={"Name": ["Tags"]},
                ),
            ],
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

import cfnlint.data.schemas.extensions.aws_codebuild_project
from cfnlint.jsonschema import ValidationError
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails


class ProjectS3Location(CfnLintJsonSchema):
    id = "E3636"
    shortdesc = "Validate CodeBuild projects using S3 also have Location"
    description = "When using 'S3' for 'Type' then you must also specify " "'Location'"
    tags = ["resources", "codebuild"]

    def __init__(self) -> None:
        super().__init__(
            keywords=[
                "Resources/AWS::CodeBuild::Project/Properties/Artifacts",
                "Resources/AWS::CodeBuild::Project/Properties/Source",
            ],
            schema_details=SchemaDetails(
                module=cfnlint.data.schemas.extensions.aws_codebuild_project,
                filename="s3_locations.json",
            ),
        )

    def message(self, instance: Any, err: ValidationError) -> str:
        return f"{err.message} when using 'Type' of 'S3'"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from typing import Any

from cfnlint.jsonschema import ValidationError, Validator
from cfnlint.rules.jsonschema.CfnLintKeyword import CfnLintKeyword
from cfnlint.schema.resolver import RefResolutionError


class ReadOnly(CfnLintKeyword):
    """Check Base Resource Configuration"""

    id = "E3040"
    shortdesc = "Validate we aren't configuring read only properties"
    description = (
        "Read only properties can be configured in a CloudFormation template but "
        "they aren't sent to the resource provider code and can cause drift."
    )
    source_url = "https://docs.aws.amazon.com/cloudformation-cli/latest/userguide/resource-type-schema.html#schema-properties-readonlyproperties"
    tags = ["resources", "properties"]

    def __init__(self):
        """Init"""
        super().__init__(keywords=["*"])

    def validate(self, validator: Validator, _: Any, instance: Any, schema: Any):
        if len(validator.context.path.cfn_path) < 3:
            return

        parent, _, section = list(validator.context.path.cfn_path)[0:3]
        if parent != "Resources" or section != "Properties":
            return

        p = "/properties/" + "/".join(list(validator.context.path.cfn_path)[3:])

        try:
            if p in validator.resolver.resolve_from_url("#/readOnlyProperties"):
                yield ValidationError(
                    "Read only properties are not allowed "
                    f"({validator.context.path.cfn_path[-1]!r})"
                )

        except RefResolutionError:
            pass



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

import logging
from collections import deque
from ipaddress import IPv4Network, IPv6Network, ip_network
from typing import Any, Iterator

from cfnlint.context import Path
from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules.helpers import get_value_from_path
from cfnlint.rules.jsonschema.CfnLintKeyword import CfnLintKeyword

LOGGER = logging.getLogger(__name__)


class VpcSubnetCidr(CfnLintKeyword):
    id = "E3059"
    shortdesc = "Validate subnet CIDRs are within the CIDRs of the VPC"
    description = (
        "When specifying subnet CIDRs for a VPC the subnet CIDRs "
        "most be within the VPC CIDRs"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-ec2-subnet.html"
    tags = ["resources", "ec2", "vpc", "subnet"]

    def __init__(self) -> None:
        super().__init__(
            keywords=[
                "Resources/AWS::EC2::VPC/Properties",
            ],
        )

    def _validate_subnets(
        self,
        source: IPv4Network | IPv6Network,
        destination: IPv4Network | IPv6Network,
    ) -> bool:
        if isinstance(source, IPv4Network) and isinstance(destination, IPv4Network):
            if source.subnet_of(destination):
                return True
            return False
        elif isinstance(source, IPv6Network) and isinstance(destination, IPv6Network):
            if source.subnet_of(destination):
                return True
            return False
        return False

    def _create_network(self, cidr: Any) -> IPv4Network | IPv6Network | None:
        if not isinstance(cidr, str):
            return None

        try:
            return ip_network(cidr)
        except Exception as e:
            LOGGER.debug(f"Unable to create network from {cidr}", e)

        return None

    def _get_vpc_cidrs(
        self, validator: Validator, instance: dict[str, Any]
    ) -> Iterator[tuple[IPv4Network | IPv6Network | None, Validator]]:
        for key in [
            "Ipv4IpamPoolId",
            "Ipv6IpamPoolId",
            "Ipv6Pool",
            "AmazonProvidedIpv6CidrBlock",
        ]:
            for value, value_validator in get_value_from_path(
                validator,
                instance,
                deque([key]),
            ):
                if value is None:
                    continue
                yield None, value_validator

        for key in ["CidrBlock", "Ipv6CidrBlock"]:
            for cidr, cidr_validator in get_value_from_path(
                validator,
                instance,
                deque([key]),
            ):

                if cidr is None:
                    continue
                yield self._create_network(cidr), cidr_validator

    def validate(
        self, validator: Validator, keywords: Any, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:

        if not validator.cfn.graph:
            return

        vpc_ipv4_networks: list[IPv4Network] = []
        vpc_ipv6_networks: list[IPv6Network] = []
        for vpc_network, _ in self._get_vpc_cidrs(validator, instance):
            if not vpc_network:
                return
            if isinstance(vpc_network, IPv4Network):
                vpc_ipv4_networks.append(vpc_network)
            # you can't specify IPV6 networks on a VPC

        template_validator = validator.evolve(
            context=validator.context.evolve(path=Path())
        )

        # dynamic vpc network (using IPAM or AWS provided)
        # allows to validate subnet overlapping even if using
        # dynamic networks
        has_dynamic_network = False

        for source, _ in validator.cfn.graph.graph.in_edges(
            validator.context.path.path[1]
        ):
            if (
                validator.cfn.graph.graph.nodes[source].get("resource_type")
                == "AWS::EC2::VPCCidrBlock"
            ):
                for cidr_props, cidr_validator in get_value_from_path(
                    template_validator,
                    validator.cfn.template,
                    deque(["Resources", source, "Properties"]),
                ):
                    for cidr_network, _ in self._get_vpc_cidrs(
                        cidr_validator, cidr_props
                    ):
                        if not cidr_network:
                            has_dynamic_network = True
                            continue
                        if isinstance(cidr_network, IPv4Network):
                            vpc_ipv4_networks.append(cidr_network)
                        else:
                            vpc_ipv6_networks.append(cidr_network)

        subnets: list[tuple[IPv4Network | IPv6Network, deque]] = []
        for source, _ in validator.cfn.graph.graph.in_edges(
            validator.context.path.path[1]
        ):
            if (
                validator.cfn.graph.graph.nodes[source].get("resource_type")
                == "AWS::EC2::Subnet"
            ):
                for subnet_props, source_validator in get_value_from_path(
                    template_validator,
                    validator.cfn.template,
                    deque(["Resources", source, "Properties"]),
                ):
                    for subnet_network, subnet_validator in self._get_vpc_cidrs(
                        source_validator, subnet_props
                    ):
                        if not subnet_network:
                            continue

                        subnets.append(
                            (subnet_network, subnet_validator.context.path.path)
                        )
                        if has_dynamic_network:
                            continue
                        if not any(
                            self._validate_subnets(
                                subnet_network,
                                vpc_network,
                            )
                            for vpc_network in vpc_ipv4_networks + vpc_ipv6_networks
                        ):
                            if isinstance(subnet_network, IPv4Network):
                                # Every VPC has to have a ipv4 network
                                # we continue if there isn't one
                                if not vpc_ipv4_networks:
                                    continue
                                reprs = (
                                    "is not a valid subnet of "
                                    f"{[f'{str(v)}' for v in vpc_ipv4_networks]!r}"
                                )
                            else:
                                if not vpc_ipv6_networks:
                                    reprs = (
                                        "is specified on a VPC that has "
                                        "no ipv6 networks defined"
                                    )
                                else:
                                    reprs = (
                                        "is not a valid subnet of "
                                        f"{[f'{str(v)}' for v in vpc_ipv6_networks]!r}"
                                    )
                            yield ValidationError(
                                (f"{str(subnet_network)!r} {reprs}"),
                                rule=self,
                                path_override=subnet_validator.context.path.path,
                            )
                            continue



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from typing import Any

from cfnlint.helpers import TRANSFORM_SAM
from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules.jsonschema import CfnLintKeyword


class ServerlessTransform(CfnLintKeyword):
    """Check if Serverless Resources exist without the Serverless Transform"""

    id = "E3038"
    shortdesc = "Check if Serverless Resources have Serverless Transform"
    description = (
        "Check that a template with Serverless Resources also includes the Serverless"
        " Transform"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/transform-aws-serverless.html"
    tags = ["resources", "transform"]

    def __init__(self) -> None:
        super().__init__(keywords=["Resources/*/Type"])

    def validate(
        self, validator: Validator, s: Any, instance: Any, schema: Any
    ) -> ValidationResult:
        if not validator.is_type(instance, "string"):
            return

        if validator.cfn.has_serverless_transform():
            return

        if instance.startswith("AWS::Serverless::"):
            yield ValidationError(
                (
                    f"{instance!r} type used without the "
                    f"serverless transform {TRANSFORM_SAM!r}"
                ),
                rule=self,
            )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from collections import deque
from typing import Any, Iterator

from cfnlint.helpers import ensure_list, is_function
from cfnlint.jsonschema import ValidationError, ValidationResult
from cfnlint.jsonschema.protocols import Validator
from cfnlint.rules.helpers import get_resource_by_name, get_value_from_path
from cfnlint.rules.jsonschema.CfnLintKeyword import CfnLintKeyword


class ServiceDynamicPorts(CfnLintKeyword):
    id = "E3049"
    shortdesc = (
        "Validate ECS tasks with dynamic host port have traffic-port ELB target groups"
    )
    description = (
        "When using an ECS task definition of host port 0 and "
        "associating that container to an ELB the target group "
        "has to have a 'HealthCheckPort' of 'traffic-port'"
    )
    tags = ["resources"]

    def __init__(self) -> None:
        super().__init__(
            keywords=["Resources/AWS::ECS::Service/Properties"],
        )

    def _filter_resource_name(self, instance: Any) -> str | None:
        fn_k, fn_v = is_function(instance)
        if fn_k is None:
            return None
        if fn_k == "Ref":
            if isinstance(fn_v, str):
                return fn_v
        elif fn_k == "Fn::GetAtt":
            name = ensure_list(fn_v)[0].split(".")[0]
            if isinstance(name, str):
                return name
        return None

    def _get_service_lb_containers(
        self, validator: Validator, instance: Any
    ) -> Iterator[tuple[str, str, str, Validator]]:
        for load_balancer, load_balancer_validator in get_value_from_path(
            validator,
            instance,
            path=deque(["LoadBalancers", "*"]),
        ):
            for name, name_validator in get_value_from_path(
                load_balancer_validator,
                load_balancer,
                path=deque(["ContainerName"]),
            ):
                if name is None or not isinstance(name, str):
                    continue
                for port, port_validator in get_value_from_path(
                    name_validator,
                    load_balancer,
                    path=deque(["ContainerPort"]),
                ):
                    if port is None or not isinstance(port, (str, int)):
                        continue
                    for tg, tg_validator in get_value_from_path(
                        port_validator,
                        load_balancer,
                        path=deque(["TargetGroupArn"]),
                    ):
                        tg = self._filter_resource_name(tg)
                        if tg is None:
                            continue
                        yield name, port, tg, tg_validator

    def _get_service_properties(
        self, validator: Validator, instance: Any
    ) -> Iterator[tuple[str, str, str, str, Validator]]:
        for task_definition_id, task_definition_validator in get_value_from_path(
            validator, instance, deque(["TaskDefinition"])
        ):
            task_definition_resource_name = self._filter_resource_name(
                task_definition_id
            )
            if task_definition_resource_name is None:
                continue
            for (
                container_name,
                container_port,
                target_group,
                lb_validator,
            ) in self._get_service_lb_containers(task_definition_validator, instance):
                yield (
                    task_definition_resource_name,
                    container_name,
                    container_port,
                    target_group,
                    lb_validator,
                )

    def _get_target_group_properties(
        self, validator: Validator, resource_name: Any
    ) -> Iterator[tuple[str | int | None, Validator]]:
        target_group, target_group_validator = get_resource_by_name(
            validator, resource_name, ["AWS::ElasticLoadBalancingV2::TargetGroup"]
        )
        if not target_group:
            return

        for port, port_validator in get_value_from_path(
            target_group_validator,
            target_group,
            path=deque(["Properties", "HealthCheckPort"]),
        ):
            if port is None:
                yield port, port_validator
                continue

            if not isinstance(port, (str, int)):
                continue
            yield port, port_validator

    def _get_task_containers(
        self,
        validator: Validator,
        resource_name: str,
        container_name: str,
        container_port: str | int,
    ) -> Iterator[Validator]:
        task_def, task_def_validator = get_resource_by_name(
            validator, resource_name, ["AWS::ECS::TaskDefinition"]
        )

        if not task_def:
            return

        for container, container_validator in get_value_from_path(
            task_def_validator,
            task_def,
            path=deque(["Properties", "ContainerDefinitions", "*"]),
        ):
            for name, name_validator in get_value_from_path(
                container_validator,
                container,
                path=deque(["Name"]),
            ):
                if not isinstance(name, str):
                    continue
                if name == container_name:
                    for host_port_validator in self._filter_port_mappings(
                        name_validator, container, container_port
                    ):
                        yield host_port_validator

    def _filter_port_mappings(
        self, validator: Validator, instance: Any, port: Any
    ) -> Iterator[Validator]:
        for port_mapping, port_mapping_validator in get_value_from_path(
            validator,
            instance,
            path=deque(["PortMappings", "*"]),
        ):
            for container_port, container_port_validator in get_value_from_path(
                port_mapping_validator,
                port_mapping,
                path=deque(["ContainerPort"]),
            ):
                if not isinstance(container_port, (str, int)):
                    continue
                if str(port) == str(container_port):
                    for host_port, host_part_validator in get_value_from_path(
                        container_port_validator,
                        port_mapping,
                        path=deque(["HostPort"]),
                    ):
                        if str(host_port) == "0":
                            yield host_part_validator

    def validate(
        self, validator: Validator, _: Any, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:

        for (
            task_definition_resource_name,
            lb_container_name,
            lb_container_port,
            lb_target_group,
            lb_service_validator,
        ) in self._get_service_properties(validator, instance):
            for container_validator in self._get_task_containers(
                lb_service_validator,
                task_definition_resource_name,
                lb_container_name,
                lb_container_port,
            ):
                for (
                    health_check_port,
                    health_check_port_validator,
                ) in self._get_target_group_properties(
                    container_validator, lb_target_group
                ):
                    if health_check_port != "traffic-port":
                        err_validator = "const"
                        if health_check_port is None:
                            err_validator = "required"
                        yield ValidationError(
                            "When using an ECS task definition of host port 0 and "
                            "associating that container to an ELB the target group "
                            "has to have a 'HealthCheckPort' of 'traffic-port'",
                            path_override=health_check_port_validator.context.path.path,
                            validator=err_validator,
                            rule=self,
                        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import pytest

from cfnlint.context import create_context_for_template
from cfnlint.jsonschema import CfnTemplateValidator, ValidationError
from cfnlint.rules.functions.DynamicReference import DynamicReference
from cfnlint.template import Template


class _TestRule:
    def validate(self, validator, s, instance, schema):
        return
        yield


@pytest.fixture(scope="module")
def rule():
    rule = DynamicReference()
    yield rule


@pytest.fixture(scope="module")
def cfn():
    return Template(
        "",
        {},
        regions=["us-east-1"],
    )


@pytest.fixture(scope="module")
def context(cfn):
    return create_context_for_template(cfn)


@pytest.mark.parametrize(
    "name,instance,schema,child_rules,expected",
    [
        (
            "Valid SSM Parameter",
            "{{resolve:ssm:Parameter}}",
            {"type": "test"},
            {
                "E1051": _TestRule(),
                "E1027": _TestRule(),
                "W1051": _TestRule(),
            },
            [],
        ),
        (
            "Valid SSM Parameter with no rules",
            "{{resolve:ssm:Parameter}}",
            {"type": "test"},
            {
                "E1051": None,
                "E1027": None,
                "W1051": None,
            },
            [],
        ),
        (
            "Valid SSM secure Parameter",
            "{{resolve:ssm-secure:Parameter}}",
            {"type": "test"},
            {
                "E1051": _TestRule(),
                "E1027": _TestRule(),
                "W1051": _TestRule(),
            },
            [],
        ),
        (
            "Valid SSM secure string with no rules",
            "{{resolve:ssm-secure:Parameter}}",
            {"type": "test"},
            {
                "E1051": None,
                "E1027": None,
                "W1051": None,
            },
            [],
        ),
        (
            "Valid SSM Parameter with integer",
            "{{resolve:ssm:Parameter:1}}",
            {"type": "test"},
            {
                "E1051": _TestRule(),
                "E1027": _TestRule(),
                "W1051": _TestRule(),
            },
            [],
        ),
        (
            "Valid when item isn't a string",
            {},
            {"type": "test"},
            {
                "E1051": _TestRule(),
                "E1027": _TestRule(),
                "W1051": _TestRule(),
            },
            [],
        ),
        (
            "Basic error when wrong length",
            "{{resolve:ssm}}",
            {"type": "test"},
            {
                "E1051": _TestRule(),
                "E1027": _TestRule(),
                "W1051": _TestRule(),
            },
            [
                ValidationError(
                    "expected minimum item count: 3, found: 2",
                    validator="minItems",
                    rule=DynamicReference(),
                )
            ],
        ),
        (
            "Invalid SSM Parameter with string version",
            "{{resolve:ssm:Parameter:a}}",
            {"type": "test"},
            {
                "E1051": _TestRule(),
                "E1027": _TestRule(),
                "W1051": _TestRule(),
            },
            [
                ValidationError(
                    "'a' does not match '\\\\d+'",
                    validator="pattern",
                    rule=DynamicReference(),
                )
            ],
        ),
        (
            "Valid secret manager",
            "{{resolve:secretsmanager:Secret}}",
            {"type": "test"},
            {
                "E1051": _TestRule(),
                "E1027": _TestRule(),
                "W1051": _TestRule(),
            },
            [],
        ),
        (
            "Valid secret manager with no child rules",
            "{{resolve:secretsmanager:Secret}}",
            {"type": "test"},
            {
                "E1051": None,
                "E1027": None,
                "W1051": None,
            },
            [],
        ),
        (
            "Valid secret manager with secretstring",
            "{{resolve:secretsmanager:Secret:SecretString}}",
            {"type": "test"},
            {
                "E1051": _TestRule(),
                "E1027": _TestRule(),
                "W1051": _TestRule(),
            },
            [],
        ),
        (
            "Valid secret manager from another account",
            "{{resolve:secretsmanager:arn:aws:secretsmanager:us-west-2:123456789012:secret:MySecret-a1b2c3:SecretString:password}}",
            {"type": "test"},
            {
                "E1051": _TestRule(),
                "E1027": _TestRule(),
                "W1051": _TestRule(),
            },
            [],
        ),
        (
            "Invalid SSM Parameter with string version",
            "{{resolve:secretsmanager:arn:aws:secretsmanager:us-east-1:012345678901:secret:my-secret:SecretString::::}}",
            {"type": "test"},
            {
                "E1051": _TestRule(),
                "E1027": _TestRule(),
                "W1051": _TestRule(),
            },
            [
                ValidationError(
                    "expected maximum item count: 13, found: 14",
                    validator="maxItems",
                    rule=DynamicReference(),
                )
            ],
        ),
    ],
)
def test_validate(name, instance, schema, child_rules, expected, rule, context, cfn):
    rule.child_rules = child_rules
    validator = CfnTemplateValidator(context=context, cfn=cfn)
    errs = list(rule.dynamicReference(validator, schema, instance, {}))
    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from test.unit.rules import BaseRuleTestCase

from cfnlint.rules.resources.CircularDependency import (
    CircularDependency,  # pylint: disable=E0401
)


class TestRulesRefCircular(BaseRuleTestCase):
    """Test Rules Get Att"""

    def setUp(self):
        """Setup"""
        super(TestRulesRefCircular, self).setUp()
        self.collection.register(CircularDependency())

    def test_file_positive(self):
        """Test Positive"""
        self.helper_file_positive()

    def test_file_negative(self):
        """Test failure"""
        err_count = 6
        self.helper_file_negative(
            "test/fixtures/templates/bad/resources_circular_dependency.yaml", err_count
        )

    def test_file_negative_fngetatt(self):
        """Test failure"""
        err_count = 9
        self.helper_file_negative(
            "test/fixtures/templates/bad/resources_circular_dependency_2.yaml",
            err_count,
        )

    def test_file_negative_dependson(self):
        """Test failure with DependsOn"""
        err_count = 2
        self.helper_file_negative(
            "test/fixtures/templates/bad/resources_circular_dependency_dependson.yaml",
            err_count,
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import pytest

from cfnlint.template import Template


@pytest.fixture
def cfn() -> Template:

    return Template(
        None,
        {
            "Parameters": {
                "Environment": {
                    "Type": "String",
                    "AllowedValues": ["dev", "test", "stage", "prod"],
                }
            },
            "Conditions": {
                "IsUsEast1": {"Fn::Equals": [{"Ref": "AWS::Region"}, "us-east-1"]},
                "IsProduction": {"Fn::Equals": [{"Ref": "Environment"}, "prod"]},
            },
            "Resources": {
                "Bucket": {"Type": "AWS::S3::Bucket", "Condition": "IsUsEast1"},
                "Vpc": {
                    "Type": "AWS::EC2::VPC",
                    "Properties": {"CidrBlock": "10.0.0.0/16"},
                },
                "SecurityGroup1": {
                    "Type": "AWS::EC2::SecurityGroup",
                    "Condition": "IsProduction",
                    "Properties": {
                        "CidrBlock": "10.0.0.0/24",
                        "VpcId": {
                            "Fn::If": ["IsUsEast1", "vpc-123", {"Ref": "AWS::NoValue"}]
                        },
                    },
                },
                "SecurityGroup2": {
                    "Type": "AWS::EC2::SecurityGroup",
                    "Condition": "IsProduction",
                    "Properties": {
                        "CidrBlock": "10.0.1.0/24",
                        "VpcId": {
                            "Fn::If": ["IsProduction", {"Ref": "Vpc"}, "vpc-abc"]
                        },
                    },
                },
            },
        },
    )


@pytest.mark.parametrize(
    "name,path,starting_conditions, expected",
    [
        (
            "Valid get path",
            ["Resources", "Vpc", "Properties", "CidrBlock"],
            {},
            [("10.0.0.0/16", {})],
        ),
        (
            "Invalid path returns nothing",
            ["Resources", "Vpc", "Properties", "DNE"],
            {},
            [],
        ),
        (
            "Short path that doesn't exist",
            ["Resources", "DNE"],
            {},
            [],
        ),
        (
            "Valid path with resource condition",
            ["Resources", "SecurityGroup1", "Properties", "CidrBlock"],
            {},
            [("10.0.0.0/24", {"IsProduction": True})],
        ),
        (
            "Valid path with resource condition",
            ["Resources", "SecurityGroup1", "Properties", "CidrBlock"],
            {"IsProduction": False},
            [],
        ),
        (
            "Valid path with resource condition with multiple conditions",
            ["Resources", "SecurityGroup1", "Properties", "VpcId"],
            {},
            [("vpc-123", {"IsProduction": True, "IsUsEast1": True})],
        ),
        (
            (
                "Valid path with resource condition with multiple "
                "conditions that conflict with each other"
            ),
            ["Resources", "SecurityGroup2", "Properties", "VpcId"],
            {},
            [({"Ref": "Vpc"}, {"IsProduction": True})],
        ),
    ],
)
def test_paths(name, path, starting_conditions, expected, cfn):

    context = cfn.context.evolve(
        conditions=cfn.context.conditions.evolve(starting_conditions)
    )

    results = list(cfn.get_cfn_path(path, context))

    assert len(results) == len(
        expected
    ), f"{name!r} test failed. Got results {results!r}"

    for i, value in enumerate(results):
        assert (
            value[0] == expected[i][0]
        ), f"{name!r} test failed for {i}. Got value {value[0]!r}"
        assert value[1].conditions.status == expected[i][1], (
            f"{name!r} test failed for {i}. Got "
            f"conditions {value[1].conditions.status[0]!r}"
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from unittest import TestCase

from cfnlint.conditions._condition import (
    ConditionAnd,
    ConditionNot,
    ConditionOr,
    ConditionUnnammed,
)
from cfnlint.conditions._utils import get_hash


class TestCondition(TestCase):
    """Test Condition"""

    def test_not_condition(self):
        """Test not condition"""

        with self.assertRaises(ValueError):
            ConditionNot(
                [
                    {"Fn::Equals": [{"Ref": "AWS::Region"}, "us-east-1"]},
                    {"Fn::Equals": [{"Ref": "AWS::Region"}, "us-west-2"]},
                ],
                {},
            )

    def test_unnamed_condition(self):
        """Test unnamed condition"""

        with self.assertRaises(ValueError):
            ConditionUnnammed(
                "equals",  # not a string
                {},
            )

    def test_not_a_list(self):
        """Test condition failures"""

        with self.assertRaises(ValueError):
            ConditionUnnammed({"Fn::And": {}}, {})

        with self.assertRaises(ValueError):
            ConditionUnnammed({"Condition": {}}, {})

        with self.assertRaises(ValueError):
            ConditionUnnammed({"BadKey": {}}, {})

    def test_condition_test(self):
        equals = {"Ref": "AWS::Region"}
        h = get_hash(equals)
        self.assertTrue(
            ConditionNot(
                [
                    {"Fn::Equals": [{"Ref": "AWS::Region"}, "us-east-1"]},
                ],
                {},
            )._test(
                {h: "us-west-2"},
            )
        )
        self.assertFalse(
            ConditionNot(
                [
                    {"Fn::Equals": [{"Ref": "AWS::Region"}, "us-east-1"]},
                ],
                {},
            )._test(
                {h: "us-east-1"},
            )
        )
        self.assertFalse(
            ConditionAnd(
                [
                    {"Fn::Equals": [{"Ref": "AWS::Region"}, "us-east-1"]},
                    {"Fn::Equals": [{"Ref": "AWS::Region"}, "us-west-2"]},
                ],
                {},
            )._test(
                {h: "us-east-1"},
            )
        )
        self.assertTrue(
            ConditionOr(
                [
                    {"Fn::Equals": ["us-east-1", {"Ref": "AWS::Region"}]},
                    {"Fn::Equals": ["us-west-2", {"Ref": "AWS::Region"}]},
                ],
                {},
            )._test(
                {h: "us-east-1"},
            )
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.rds.DbInstanceDbInstanceClassWithEngine import (
    DbInstanceDbInstanceClassWithEngine,
)


@pytest.fixture(scope="module")
def rule():
    rule = DbInstanceDbInstanceClassWithEngine()
    yield rule


@pytest.fixture
def schema():
    return {
        "allOf": [
            {
                "if": {
                    "properties": {
                        "DBInstanceClass": {"type": "string"},
                        "Engine": {"const": "mysql", "type": "string"},
                        "EngineVersion": {
                            "pattern": "^(8\\.0\\..+|8\\.0)$",
                            "type": "string",
                        },
                    },
                    "required": ["Engine", "EngineVersion", "DBInstanceClass"],
                },
                "then": {
                    "properties": {
                        "DBInstanceClass": {
                            "enum": [
                                "db.t3.small",
                            ]
                        }
                    }
                },
            }
        ]
    }


@pytest.mark.parametrize(
    "name,instance,expected",
    [
        (
            "Valid because not an object",
            [],
            [],
        ),
        (
            "Valid class",
            {
                "Engine": "mysql",
                "EngineVersion": "8.0.39",
                "DBInstanceClass": "db.t3.small",
            },
            [],
        ),
        (
            "Valid because its missing EngineVersion",
            {
                "Engine": "mysql",
                "DBInstanceClass": "db.t2.small",
            },
            [],
        ),
        (
            "Valid because its missing Engine",
            {
                "EngineVersion": "8.0.39",
                "DBInstanceClass": "db.t2.small",
            },
            [],
        ),
        (
            "Invalid DBInstanceClass for Engine and EngineVersion",
            {
                "Engine": "mysql",
                "EngineVersion": "8.0.39",
                "DBInstanceClass": "db.t2.small",
            },
            [
                ValidationError(
                    "'db.t2.small' is not one of ['db.t3.small']",
                    validator="enum",
                    rule=DbInstanceDbInstanceClassWithEngine(),
                    path=deque(["DBInstanceClass"]),
                    schema_path=deque(
                        ["allOf", 0, "then", "properties", "DBInstanceClass", "enum"]
                    ),
                )
            ],
        ),
    ],
)
def test_validate(name, instance, expected, rule, schema, validator):
    rule._schema = schema
    errors = list(rule.validate(validator, "", instance, {}))

    assert errors == expected, f"Test {name!r} got {errors!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule


class RequiredOr(CloudFormationLintRule):
    id = "E3058"
    shortdesc = "Validate at least one of the properties are required"
    description = "Make sure at least one of the resource properties are included"
    source_url = "https://github.com/aws-cloudformation/cfn-lint/blob/main/docs/cfn-schema-specification.md#requiredor"
    tags = ["resources"]



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

import regex as re

from cfnlint.helpers import REGEX_DYN_REF
from cfnlint.jsonschema import ValidationResult, Validator
from cfnlint.rules.jsonschema.CfnLintKeyword import CfnLintKeyword


class Aliases(CfnLintKeyword):
    """Check if CloudFront Aliases are valid domain names"""

    id = "E3013"
    shortdesc = "CloudFront Aliases"
    description = "CloudFront aliases should contain valid domain names"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-cloudfront-distribution-distributionconfig.html#cfn-cloudfront-distribution-distributionconfig-aliases"
    tags = ["properties", "cloudfront"]

    def __init__(self):
        """Init"""
        super().__init__(
            keywords=[
                "Resources/AWS::CloudFront::Distribution/Properties/DistributionConfig/Aliases/*"
            ]
        )

    def validate(
        self, validator: Validator, _, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:

        if isinstance(instance, str):
            if re.match(REGEX_DYN_REF, instance):
                return
            for err in validator.descend(
                instance=instance,
                schema={
                    "pattern": "^(?!.*(?:\\.\\*\\.)).*",
                },
            ):
                err.rule = self
                yield err

            for err in validator.descend(
                instance=instance,
                schema={
                    # ruff: noqa: E501
                    "pattern": "^(?:[a-z0-9\\*](?:[a-z0-9-]{0,61}[a-z0-9])?\\.)+[a-z0-9][a-z0-9-]{0,61}[a-z0-9]$",
                },
            ):
                err.rule = self
                yield err



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from typing import Any

import cfnlint.data.schemas.extensions.aws_elasticloadbalancingv2_targetgroup
from cfnlint.jsonschema import ValidationResult, Validator
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails


class TargetGroupHealthCheckProtocolRestrictions(CfnLintJsonSchema):
    id = "E3684"
    shortdesc = "Validate target group health check protocol property restrictions"
    description = (
        "When a TargetGroup health check protocol is specified there are "
        "restrictions on other properties."
    )
    tags = ["resources"]

    def __init__(self) -> None:
        super().__init__(
            keywords=[
                "Resources/AWS::ElasticLoadBalancingV2::TargetGroup/Properties",
            ],
            schema_details=SchemaDetails(
                module=cfnlint.data.schemas.extensions.aws_elasticloadbalancingv2_targetgroup,
                filename="healthcheckprotocol_restrictions.json",
            ),
            all_matches=True,
        )

    def validate(
        self, validator: Validator, keywords: Any, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:
        for err in super().validate(validator, keywords, instance, schema):
            if not err.schema:
                err.message = (
                    f"Additional properties are not allowed ({err.path[0]!r} "
                    "was unexpected)"
                )
                err.validator = "additionalProperties"
            yield err



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from test.testlib.testcase import BaseTestCase

from cfnlint.decode import cfn_yaml
from cfnlint.template.transforms._sam import Transform


class TestTransform(BaseTestCase):
    """Test Transform Parsing"""

    def test_parameter_for_autopublish_version(self):
        """Test Parameter is created for autopublish version run"""
        filename = "test/fixtures/templates/good/transform/auto_publish_alias.yaml"
        region = "us-east-1"
        template = cfn_yaml.load(filename)
        transformed_template = Transform(filename, template, region)
        transformed_template.transform_template()
        self.assertDictEqual(
            transformed_template._parameters, {"Stage1": "Alias", "Stage2": "Alias"}
        )
        self.assertDictEqual(
            transformed_template._template.get("Resources")
            .get("SkillFunctionAliasAlias")
            .get("Properties"),
            {
                "Name": "Alias",
                "FunctionName": {"Ref": "SkillFunction"},
                "FunctionVersion": {
                    "Fn::GetAtt": ["SkillFunctionVersion55ff35af87", "Version"]
                },
            },
        )

    def test_conversion_of_application_location(self):
        """Tests if serverless application converts location to string when dict"""
        filename = "test/fixtures/templates/good/transform/applications_location.yaml"
        region = "us-east-1"
        template = cfn_yaml.load(filename)
        transformed_template = Transform(filename, template, region)
        transformed_template.transform_template()
        self.assertEqual(
            transformed_template._template.get("Resources")
            .get("App1")
            .get("Properties")
            .get("TemplateURL"),
            "./step_function_local_definition.yaml",
        )
        self.assertEqual(
            transformed_template._template.get("Resources")
            .get("App2")
            .get("Properties")
            .get("TemplateURL"),
            "s3://bucket/value",
        )

    def test_conversion_of_step_function_definition_uri(self):
        """
        Tests that the a serverless step function can convert
        a local path to a s3 path
        """
        filename = (
            "test/fixtures/templates/good/transform/step_function_local_definition.yaml"
        )
        region = "us-east-1"
        template = cfn_yaml.load(filename)
        transformed_template = Transform(filename, template, region)
        transformed_template.transform_template()
        self.assertDictEqual(
            transformed_template._template.get("Resources")
            .get("StateMachine")
            .get("Properties")
            .get("DefinitionS3Location"),
            {"Bucket": "bucket", "Key": "value"},
        )

    def test_conversion_of_serverless_function_uri(self):
        """Tests that the a serverless function can convert a CodeUri Sub"""
        filename = "test/fixtures/templates/good/transform/function_use_s3_uri.yaml"
        region = "us-east-1"
        template = cfn_yaml.load(filename)
        transformed_template = Transform(filename, template, region)
        transformed_template.transform_template()
        self.assertDictEqual(
            transformed_template._template.get("Resources")
            .get("Function")
            .get("Properties")
            .get("Code"),
            {"S3Bucket": "bucket", "S3Key": "value"},
        )

    def test_resource_properties_dont_exist(self):
        filename = "test/fixtures/templates/bad/transform/no_properties.yaml"
        region = "us-east-1"
        template = cfn_yaml.load(filename)
        transformed_template = Transform(filename, template, region)
        results = transformed_template.transform_template()
        self.assertEqual(len(results), 1)

    def test_parameter_for_autopublish_version_bad(self):
        """Test Parameter is created for autopublish version run"""
        filename = "test/fixtures/templates/bad/transform/auto_publish_alias.yaml"
        region = "us-east-1"
        template = cfn_yaml.load(filename)
        transformed_template = Transform(filename, template, region)
        transformed_template.transform_template()
        self.assertDictEqual(transformed_template._parameters, {})

    def test_test_function_using_image_good(self):
        """Test Parameter is created for autopublish version run"""
        filename = "test/fixtures/templates/good/transform/function_using_image.yaml"
        region = "us-east-1"
        template = cfn_yaml.load(filename)
        transformed_template = Transform(filename, template, region)
        transformed_template.transform_template()
        self.assertDictEqual(transformed_template._parameters, {})

    def test_sam_with_language_extension(self):
        """Test language extension"""
        filename = "test/fixtures/templates/good/transform/language_extension.yaml"
        region = "us-east-1"
        template = cfn_yaml.load(filename)
        transformed_template = Transform(filename, template, region)
        results = transformed_template.transform_template()
        self.assertDictEqual(
            transformed_template._template.get("Rules"),
            {
                "IsAutPublishAliasParameterProd": {
                    "Assertions": [
                        {
                            "Assert": {
                                "Fn::Not": [{"Fn::Equals": [{"Ref": "DBPolicy"}, ""]}]
                            }
                        }
                    ],
                    "RuleCondition": {"Fn::Equals": [{"Ref": "Environment"}, "Prod"]},
                }
            },
        )
        self.assertEqual(results, [])

    def test_parameter_for_autopublish_code_sha256(self):
        filename = (
            "test/fixtures/templates/good/transform/auto_publish_code_sha256.yaml"
        )
        region = "us-east-1"
        template = cfn_yaml.load(filename)
        transformed_template = Transform(filename, template, region)
        transformed_template.transform_template()
        self.assertDictEqual(transformed_template._parameters, {})
        self.assertDictEqual(
            transformed_template._template.get("Resources")
            .get("LambdaFunction")
            .get("Properties"),
            {
                "Code": {
                    "S3Bucket": {"Ref": "CodeBucket"},
                    "S3Key": {"Fn::Sub": "${ Basepath }/lambda.zip"},
                },
                "Description": "fakesha",
                "FunctionName": {"Fn::Sub": "${ AWS::StackName }"},
                "Handler": "lambda.handler",
                "MemorySize": 256,
                "Role": {"Fn::GetAtt": ["LambdaFunctionRole", "Arn"]},
                "Runtime": "python3.12",
                "Tags": [{"Key": "lambda:createdBy", "Value": "SAM"}],
                "Timeout": 30,
                "TracingConfig": {"Mode": "PassThrough"},
            },
        )
        self.assertIn(
            "LambdaFunctionVersionfakesha",
            transformed_template._template.get("Resources").keys(),
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from typing import Any, Iterator

from cfnlint.jsonschema import ValidationError, Validator
from cfnlint.rules.functions._BaseFn import BaseFn, singular_types


class ImportValue(BaseFn):
    """Check if ImportValue values are correct"""

    id = "E1016"
    shortdesc = "ImportValue validation of parameters"
    description = "Making sure the ImportValue function is properly configured"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-importvalue.html"
    tags = ["functions", "importvalue"]

    def __init__(self) -> None:
        super().__init__(
            "Fn::ImportValue",
            singular_types,
        )
        self.child_rules = {
            "W6001": None,
        }

    def fn_importvalue(
        self, validator: Validator, s: Any, instance: Any, schema: Any
    ) -> Iterator[ValidationError]:
        errs = list(super().validate(validator, s, instance, schema))
        if errs:
            yield from iter(errs)

        for rule in self.child_rules.values():
            if rule is None or not hasattr(rule, "validate"):
                continue

            yield from rule.validate(validator, s, instance, schema)



"""
Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

import fnmatch
import json
import logging
import os
import sys
from functools import lru_cache
from typing import TYPE_CHECKING, Any

import jsonpatch

from cfnlint.helpers import load_resource
from cfnlint.schema._exceptions import SchemaNotFoundError

if TYPE_CHECKING:
    from cfnlint.schema._patch import SchemaPatch

LOGGER = logging.getLogger(__name__)


class _FileLocation:
    def __init__(self, path: list[str]):
        self.path_relative = os.path.join(
            os.path.dirname(__file__),
            "..",
            *path,
        )
        self.module = ".".join(["cfnlint"] + path[:])
        self.path = path


class OtherSchemaManager:
    def __init__(self) -> None:
        self._root = _FileLocation(
            [
                "data",
                "schemas",
            ]
        )

        self._schemas: dict[str, dict[str, Any]] = {}
        self._schema_modules: dict[str, Any] = {}
        self.reset()

    def reset(self) -> None:
        """
        Reset's the cache so specs can be reloaded.
        Important function when processing many templates
        and using spec patching
        """
        self._schemas = {}
        self.get_schema.cache_clear()

    def load_registry_schemas(self, path: str) -> None:
        """Load extra registry schemas from a directory

        Args:
            path (str): the directory to load schema files from
        Returns:
            None: None
        """
        for dirpath, _, filenames in os.walk(path):
            for filename in fnmatch.filter(filenames, "*.json"):
                with open(os.path.join(dirpath, filename), "r", encoding="utf-8") as fh:
                    schema: dict[str, dict[str, Any]] = json.load(fh)
                    self._schemas[path] = schema

    @lru_cache(maxsize=None)
    def get_schema(self, path: str) -> dict[str, Any]:
        """Get the provider resource schema and cache it to speed up future lookups

        Args:
            path: the path of the schema using the dot notation
        Returns:
            dict: returns the schema
        """

        schema = self._schemas.get(path)
        if schema is not None:
            return schema

        # dynamically import the modules as needed
        path_parts = path.split(".")
        module_name = ".".join(path_parts[:-1])
        filename = f"{path_parts[-1]}.json"
        try:
            self._schema_modules[module_name] = __import__(
                f"{self._root.module}.{module_name}", fromlist=[""]
            )

            self._schemas[path] = load_resource(
                self._schema_modules[module_name],
                filename=filename,
            )
            return self._schemas[path]
        except Exception as e:
            raise SchemaNotFoundError(path) from e

    def patch(self, patch: SchemaPatch, region: str) -> None:
        """Patch the schemas as needed

        Args:
            patch: The patches to be applied to the schemas
        Returns:
            None: Returns when completed
        """

        for schema_path, patches in patch.patches.items():
            try:
                self.get_schema(path=schema_path)
            except SchemaNotFoundError:
                # Resource type doesn't exist in this region
                continue
            try:
                jsonpatch.JsonPatch(patches).apply(
                    self._schemas[schema_path], in_place=True
                )
            except Exception as e:
                print(f"Error applying patch {patches} for {schema_path}: {e}")
                sys.exit(1)


OTHER_SCHEMA_MANAGER: OtherSchemaManager = OtherSchemaManager()



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.helpers import FUNCTIONS_SINGLE
from cfnlint.rules.functions._BaseFn import BaseFn


class Base64(BaseFn):
    """Check if Base64 values are correct"""

    id = "E1021"
    shortdesc = "Base64 validation of parameters"
    description = "Making sure the Base64 function is properly configured"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-base64.html"
    tags = ["functions", "base64"]

    def __init__(self) -> None:
        super().__init__(
            "Fn::Base64",
            ("string",),
            tuple(FUNCTIONS_SINGLE),
        )
        self.fn_base64 = self.validate



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from test.testlib.testcase import BaseTestCase

from cfnlint import ConfigMixIn
from cfnlint.config import _DEFAULT_RULESDIR
from cfnlint.rules import Rules
from cfnlint.runner import Runner
from cfnlint.schema import PROVIDER_SCHEMA_MANAGER, SchemaPatch, reset


class TestOverrideRequired(BaseTestCase):
    """Used for Testing Rules"""

    def setUp(self):
        """Setup"""
        self.collection = Rules.create_from_directory(_DEFAULT_RULESDIR)
        self.region = "us-east-1"

    def tearDown(self):
        """Tear Down"""
        # Reset the Spec override to prevent other tests to fail
        reset()

    def test_success_run(self):
        """Success test"""
        filename = "test/fixtures/templates/good/override/required.yaml"

        PROVIDER_SCHEMA_MANAGER.patch(
            SchemaPatch(
                included_resource_types=[],
                excluded_resource_types=[],
                patches={
                    "AWS::S3::Bucket": [
                        {"op": "add", "path": "/required", "value": ["BucketName"]}
                    ]
                },
            ),
            region=self.region,
        )

        config = ConfigMixIn(
            regions=[self.region],
            templates=[filename],
        )
        runner = Runner(config)
        runner.rules = self.collection

        self.assertEqual([], list(runner.run()))

    def test_fail_run(self):
        """Failure test required"""
        filename = "test/fixtures/templates/bad/override/required.yaml"

        PROVIDER_SCHEMA_MANAGER.patch(
            SchemaPatch(
                included_resource_types=[],
                excluded_resource_types=[],
                patches={
                    "AWS::S3::Bucket": [
                        {"op": "add", "path": "/required", "value": ["BucketName"]}
                    ]
                },
            ),
            region=self.region,
        )

        config = ConfigMixIn(
            regions=[self.region],
            templates=[filename],
            ignore_checks=["I", "W", "E"],
            mandatory_checks=["E3003"],
        )
        runner = Runner(config)
        runner.rules = self.collection

        errs = list(runner.run())
        self.assertEqual(1, len(errs), errs)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0

# Code is taken from jsonschema package and adapted CloudFormation use
# https://github.com/yaml/pyyaml/blob/a2d19c0234866dc9d4d55abf3009699c258bb72f/lib/yaml/scanner.py#L46
"""

import fileinput
import logging
import sys

from yaml import MappingNode, ScalarNode, SequenceNode
from yaml.composer import Composer
from yaml.constructor import ConstructorError, SafeConstructor
from yaml.reader import Reader
from yaml.resolver import Resolver
from yaml.scanner import Scanner

from cfnlint.decode.mark import Mark
from cfnlint.decode.node import dict_node, list_node, str_node
from cfnlint.rules import Match
from cfnlint.rules.errors import ParseError

try:
    from yaml._yaml import CParser as Parser  # pylint: disable=ungrouped-imports,

    cyaml = True
except ImportError:
    from yaml.parser import Parser  # type: ignore # pylint: disable=ungrouped-imports

    cyaml = False

UNCONVERTED_SUFFIXES = ["Ref", "Condition"]
FN_PREFIX = "Fn::"

LOGGER = logging.getLogger(__name__)


class CfnParseError(ConstructorError):
    """
    Error thrown when the template contains Cfn Error
    """

    def __init__(self, filename, errors):
        if isinstance(errors, Match):
            errors = [errors]

        # Call the base class constructor with the parameters it needs
        super().__init__(errors[0].message)

        # Now for your custom code...
        self.filename = filename
        self.matches = errors


def build_match(filename, message, line_number, column_number, key):
    return Match.create(
        message=message,
        filename=filename,
        rule=ParseError(),
        linenumber=line_number + 1,
        columnnumber=column_number + 1,
        linenumberend=line_number + 1,
        columnnumberend=column_number + 1 + len(key),
    )


class NodeConstructor(SafeConstructor):
    """
    Node Constructors for loading different types in Yaml
    """

    def __init__(self, filename):
        # Call the base class constructor
        super().__init__()
        self.filename = filename

    def flatten_mapping(self, node):
        # Rewrote to handle merging and overwriting keys
        if any(key_node.tag == "tag:yaml.org,2002:merge" for key_node, _ in node.value):
            super().flatten_mapping(node)
            setattr(node, "using_merge", True)
        super().flatten_mapping(node)

    # To support lazy loading, the original constructors first yield
    # an empty object, then fill them in when iterated. Due to
    # laziness we omit this behaviour (and will only do "deep
    # construction") by first exhausting iterators, then yielding
    # copies.
    def construct_yaml_map(self, node):
        # Check for duplicate keys on the current level, this is not desirable
        # because a dict does not support this. It overwrites it with the last
        # occurance, which can give unexpected results
        mapping = {}
        self.flatten_mapping(node)
        matches = []
        for key_node, value_node in node.value:
            key = self.construct_object(key_node, False)
            value = self.construct_object(value_node, False)

            if key is None:
                raise CfnParseError(
                    self.filename,
                    [
                        build_match(
                            filename=self.filename,
                            message=(
                                f"Null key {key_node.value!r} not supported "
                                f"(line {key_node.start_mark.line + 1})"
                            ),
                            line_number=key_node.start_mark.line,
                            column_number=key_node.start_mark.column,
                            key=key_node.value,
                        ),
                    ],
                )
            if not getattr(node, "using_merge", False):
                for key_dup in mapping:
                    if key_dup == key:
                        if matches:
                            matches.append(
                                build_match(
                                    filename=self.filename,
                                    message=(
                                        f"Duplicate found {key!r} (line"
                                        f" {key_node.start_mark.line + 1})"
                                    ),
                                    line_number=key_node.start_mark.line,
                                    column_number=key_node.start_mark.column,
                                    key=key,
                                )
                            )
                        else:
                            matches.extend(
                                [
                                    build_match(
                                        filename=self.filename,
                                        message=(
                                            f"Duplicate found {key!r} (line"
                                            f" {key_dup.start_mark.line + 1})"
                                        ),
                                        line_number=key_dup.start_mark.line,
                                        column_number=key_dup.start_mark.column,
                                        key=key,
                                    ),
                                    build_match(
                                        filename=self.filename,
                                        message=(
                                            f"Duplicate found {key!r} (line"
                                            f" {key_node.start_mark.line + 1})"
                                        ),
                                        line_number=key_node.start_mark.line,
                                        column_number=key_node.start_mark.column,
                                        key=key,
                                    ),
                                ],
                            )
            try:
                mapping[key] = value
            except Exception as exc:
                raise CfnParseError(
                    self.filename,
                    [
                        build_match(
                            filename=self.filename,
                            message=(
                                f'Unhashable type "{key}" (line'
                                f" {key.start_mark.line + 1})"
                            ),
                            line_number=key.start_mark.line,
                            column_number=key.start_mark.column,
                            key=key,
                        ),
                    ],
                ) from exc

        if matches:
            raise CfnParseError(
                self.filename,
                matches,
            )

        (obj,) = SafeConstructor.construct_yaml_map(self, node)

        using_merge = False if not hasattr(node, "using_merge") else node.using_merge
        return dict_node(obj, node.start_mark, node.end_mark, using_merge)

    def construct_yaml_str(self, node):
        obj = SafeConstructor.construct_yaml_str(self, node)
        assert isinstance(obj, (str))
        return str_node(obj, node.start_mark, node.end_mark)

    def construct_yaml_seq(self, node):
        (obj,) = SafeConstructor.construct_yaml_seq(self, node)
        assert isinstance(obj, list)
        return list_node(obj, node.start_mark, node.end_mark)


NodeConstructor.add_constructor(  # type: ignore
    "tag:yaml.org,2002:map", NodeConstructor.construct_yaml_map
)

NodeConstructor.add_constructor(  # type: ignore
    "tag:yaml.org,2002:str", NodeConstructor.construct_yaml_str
)

NodeConstructor.add_constructor(  # type: ignore
    "tag:yaml.org,2002:seq", NodeConstructor.construct_yaml_seq
)


class _Scanner(Scanner):

    def __init__(self) -> None:
        super().__init__()
        self.ESCAPE_REPLACEMENTS = {
            "0": "\0",
            "a": "\x07",
            "b": "\x08",
            "t": "\x09",
            "\t": "\x09",
            "n": "\x0A",
            "v": "\x0B",
            "f": "\x0C",
            "r": "\x0D",
            "e": "\x1B",
            " ": "\x20",
            '"': '"',
            "\\": "\\",
            "N": "\x85",
            "_": "\xA0",
            "L": "\u2028",
            "P": "\u2029",
        }


# pylint: disable=too-many-ancestors
class MarkedLoader(Reader, _Scanner, Parser, Composer, NodeConstructor, Resolver):
    """
    Class for marked loading YAML
    """

    # pylint: disable=non-parent-init-called,super-init-not-called

    def __init__(self, stream, filename):
        Reader.__init__(self, stream)
        _Scanner.__init__(self)
        if cyaml:
            Parser.__init__(self, stream)
        else:
            Parser.__init__(self)
        Composer.__init__(self)
        SafeConstructor.__init__(self)
        Resolver.__init__(self)
        NodeConstructor.__init__(self, filename)

    def construct_getatt(self, node):
        """
        Reconstruct !GetAtt into a list
        """

        if isinstance(node.value, (str)):
            return list_node(node.value.split(".", 1), node.start_mark, node.end_mark)
        if isinstance(node.value, list):
            if any(isinstance(value, tuple) for value in node.value):
                return dict_node(
                    self.construct_object(
                        MappingNode(
                            tag="tag:yaml.org,2002:map",
                            value=node.value,
                            start_mark=node.start_mark,
                            end_mark=node.end_mark,
                        ),
                        deep=True,
                    )
                )
            else:
                return list_node(
                    [self.construct_object(child, deep=True) for child in node.value],
                    node.start_mark,
                    node.end_mark,
                )

        raise ValueError(f"Unexpected GetAtt format: {type(node.value)}")


def multi_constructor(loader, tag_suffix, node):
    """
    Deal with !Ref style function format
    """

    if tag_suffix not in UNCONVERTED_SUFFIXES:
        tag_suffix = f"{FN_PREFIX}{tag_suffix}"

    constructor = None
    if tag_suffix == "Fn::GetAtt":
        constructor = loader.construct_getatt
    elif isinstance(node, ScalarNode):
        constructor = loader.construct_scalar
    elif isinstance(node, SequenceNode):
        constructor = loader.construct_sequence
    elif isinstance(node, MappingNode):
        constructor = loader.construct_mapping
    else:
        raise f"Bad tag: !{tag_suffix}"

    return dict_node({tag_suffix: constructor(node)}, node.start_mark, node.end_mark)


def loads(yaml_string, fname=None):
    """
    Load the given YAML string
    """
    loader = MarkedLoader(yaml_string, fname)
    loader.add_multi_constructor("!", multi_constructor)
    template = loader.get_single_data()
    # Convert an empty file to an empty dict
    if template is None:
        template = dict_node({}, Mark(0, 0), Mark(0, 0))

    return template


def load(filename):
    """
    Load the given YAML file
    """

    content = ""

    if (filename is None) and (not sys.stdin.isatty()):
        filename = "-"  # no filename provided, it's stdin
        fileinput_args = {"files": filename}
        if sys.version_info.major <= 3 and sys.version_info.minor >= 10:
            fileinput_args["encoding"] = "utf-8"
        with fileinput.input(**fileinput_args) as f:
            for line in f:
                content = content + line
    else:
        with open(filename, encoding="utf-8") as fp:
            content = fp.read()

    return loads(content, filename)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError

# ruff: noqa: E501
from cfnlint.rules.resources.elasticloadbalancingv2.TargetGroupTargetTypeRestrictions import (
    TargetGroupTargetTypeRestrictions,
)


@pytest.fixture(scope="module")
def rule():
    rule = TargetGroupTargetTypeRestrictions()
    yield rule


@pytest.mark.parametrize(
    "instance,expected",
    [
        (
            {"TargetType": "lambda"},
            [],
        ),
        (
            {"TargetType": "ip", "Port": "", "Protocol": "", "VpcId": ""},
            [],
        ),
        (
            [],
            [],
        ),
        (
            {"TargetType": "lambda"},
            [],
        ),
        (
            {"TargetType": "lambda", "Port": "443"},
            [
                ValidationError(
                    "Additional properties are not allowed ('Port' was unexpected)",
                    rule=TargetGroupTargetTypeRestrictions(),
                    path=deque(["Port"]),
                    validator="additionalProperties",
                    schema_path=deque(["then", "then", "properties", "Port"]),
                )
            ],
        ),
        (
            {
                "TargetType": "ip",
            },
            [
                ValidationError(
                    "'Port' is a required property",
                    rule=TargetGroupTargetTypeRestrictions(),
                    path=deque([]),
                    validator="required",
                    schema_path=deque(["then", "else", "required"]),
                ),
                ValidationError(
                    "'Protocol' is a required property",
                    rule=TargetGroupTargetTypeRestrictions(),
                    path=deque([]),
                    validator="required",
                    schema_path=deque(["then", "else", "required"]),
                ),
                ValidationError(
                    "'VpcId' is a required property",
                    rule=TargetGroupTargetTypeRestrictions(),
                    path=deque([]),
                    validator="required",
                    schema_path=deque(["then", "else", "required"]),
                ),
            ],
        ),
    ],
)
def test_backup_lifecycle(instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))
    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from collections import deque
from typing import Any

from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules.jsonschema.CfnLintKeyword import CfnLintKeyword


class RecordSetName(CfnLintKeyword):
    """Check if a Route53 Resoruce Records Name is valid with a HostedZoneName"""

    id = "E3041"
    shortdesc = "RecordSet HostedZoneName is a superdomain of or equal to Name"
    description = (
        "In a RecordSet, the HostedZoneName must be a superdomain of or equal to"
        " the Name being validated"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-route53-recordset.html#cfn-route53-recordset-name"
    tags = ["resource", "properties", "route53"]

    def __init__(self):
        """Init"""
        super().__init__(["Resources/AWS::Route53::RecordSet/Properties"])

    def validate(
        self, validator: Validator, keywords: Any, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:
        property_sets = validator.cfn.get_object_without_conditions(
            instance, ["Name", "HostedZoneName"]
        )
        for property_set in property_sets:
            props = property_set.get("Object")
            name = props.get("Name", None)
            hz_name = props.get("HostedZoneName", None)
            if isinstance(name, str) and isinstance(hz_name, str):
                if not hz_name.endswith("."):
                    yield ValidationError(
                        f"{hz_name!r} must end in a dot",
                        path=deque(["HostedZoneName"]),
                        instance=props.get("HostedZoneName"),
                    )
                    hz_name = f"{hz_name}."

                if not name.endswith("."):
                    name = f"{name}."

                if name != hz_name and not name.endswith(f".{hz_name}"):
                    yield ValidationError(
                        (
                            f"{props.get('Name')!r} must be a subdomain "
                            f"of or equal to {props.get('HostedZoneName')!r}"
                        ),
                        path=deque(["Name"]),
                        instance=props.get("Name"),
                    )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.context import Path
from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.RetentionPeriodOnResourceTypesWithAutoExpiringContent import (  # noqa: E501
    RetentionPeriodOnResourceTypesWithAutoExpiringContent,
)


@pytest.fixture(scope="module")
def rule():
    rule = RetentionPeriodOnResourceTypesWithAutoExpiringContent()
    yield rule


@pytest.fixture
def template():
    return {
        "Resources": {
            "MyRdsDbInstance": {
                "Type": "AWS::RDS::DBInstance",
            }
        }
    }


@pytest.fixture
def path():
    return Path(
        path=deque(["Resources", "MyRdsDbInstance", "Properties"]),
        cfn_path=deque(["Resources", "AWS::RDS::DBInstance", "Properties"]),
    )


@pytest.mark.parametrize(
    "name,instance,expected",
    [
        (
            "Valid retention period",
            {"MessageRetentionPeriod": "90"},
            [],
        ),
        (
            "Valid with SourceDBInstanceIdentifier",
            {"SourceDBInstanceIdentifier": "source-db"},
            [],
        ),
        (
            "Invalid when not specified",
            {
                "Engine": "mysql",
            },
            [
                ValidationError(
                    (
                        "'BackupRetentionPeriod' is a required property (The "
                        "default retention period will delete the data after "
                        "a pre-defined time. Set an explicit values to avoid "
                        "data loss on resource)"
                    ),
                    rule=RetentionPeriodOnResourceTypesWithAutoExpiringContent(),
                    schema_path=deque(["then", "required"]),
                    validator="required",
                    validator_value=["BackupRetentionPeriod"],
                )
            ],
        ),
    ],
)
def test_validate(name, instance, expected, rule, validator):
    errs = list(rule.validate(validator, False, instance, {}))

    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from collections import deque
from typing import Any

from cfnlint.jsonschema import ValidationResult, Validator
from cfnlint.jsonschema.exceptions import ValidationError
from cfnlint.rules.helpers import get_value_from_path
from cfnlint.rules.jsonschema.CfnLintKeyword import CfnLintKeyword


class PipelineActionConfiguration(CfnLintKeyword):
    id = "E3703"
    shortdesc = "Validate the configuration of a pipeline action"
    description = (
        "When definition a CodePipeline certain action types have "
        "configuration constraints so this rule validates them"
    )
    tags = ["resources", "codepipeline"]

    def __init__(self) -> None:
        super().__init__(
            keywords=[
                "Resources/AWS::CodePipeline::Pipeline/Properties/Stages/*/Actions/*",
            ],
        )

    def validate(
        self, validator: Validator, keywords: Any, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:
        if not validator.is_type(instance, "object"):
            return

        for template_path, template_path_validator in get_value_from_path(
            validator, instance, path=deque(["Configuration", "TemplatePath"])
        ):
            if not validator.is_type(template_path, "string"):
                continue
            input_artficat_names = set()
            for input_name, _ in get_value_from_path(
                template_path_validator,
                instance,
                path=deque(["InputArtifacts", "*", "Name"]),
            ):
                if not validator.is_type(input_name, "string"):
                    continue
                if template_path.startswith(f"{input_name}::"):
                    break
                input_artficat_names.add(input_name)
            else:
                yield ValidationError(
                    (
                        f"{template_path.split('::')[0]!r} is not "
                        f"one of {list(input_artficat_names)!r}"
                    ),
                    validator="enum",
                    rule=self,
                    path_override=template_path_validator.context.path.path,
                )

        for role, role_validator in get_value_from_path(
            validator, instance, path=deque(["Configuration", "RoleArn"])
        ):
            if role is None:
                continue
            role_validator = role_validator.evolve(
                schema={"format": "AWS::IAM::Role.Arn"}
            )

            for err in role_validator.iter_errors(role):
                err.rule = self
                err.path_override = role_validator.context.path.path
                yield err



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from typing import Any

import cfnlint.data.schemas.extensions.aws_route53_recordset
from cfnlint.jsonschema.exceptions import ValidationError
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails


class RecordSetAlias(CfnLintJsonSchema):
    """Check Route53 Recordset Configuration"""

    id = "E3029"
    shortdesc = "Validate Route53 record set aliases"
    description = (
        "When using alias records you can't specify TTL or certain types are allowed"
    )
    source_url = "https://docs.aws.amazon.com/Route53/latest/DeveloperGuide/ResourceRecordTypes.html"
    tags = ["resources", "route53", "record_set"]

    def __init__(self) -> None:
        super().__init__(
            keywords=[
                "Resources/AWS::Route53::RecordSet/Properties",
                "Resources/AWS::Route53::RecordSetGroup/Properties/RecordSets/*",
            ],
            schema_details=SchemaDetails(
                module=cfnlint.data.schemas.extensions.aws_route53_recordset,
                filename="recordset_alias.json",
            ),
        )

    def message(self, instance: Any, err: ValidationError) -> str:
        return "Additional properties are not allowed ('TTL' was unexpected)"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema.validators import ValidationError
from cfnlint.rules.resources.stepfunctions.StateMachineDefinition import (
    StateMachineDefinition,
)


@pytest.fixture(scope="module")
def rule():
    rule = StateMachineDefinition()
    yield rule


# These state machines for testing have long lines
# with lots of indentation
# ruff: noqa: E501


@pytest.mark.parametrize(
    "name,instance,expected",
    [
        (
            "Valid string json",
            {
                "Definition": {
                    "Comment": (
                        "An example of the Amazon States "
                        "Language for notification on an "
                        "AWS Batch job completion"
                    ),
                    "StartAt": "Submit Batch Job",
                    "TimeoutSeconds": 3600,
                    "States": {
                        "Submit Batch Job": {
                            "Type": "Task",
                            "Resource": "arn:aws:states:::batch:submitJob.sync",
                            "Parameters": {
                                "JobName": "BatchJobNotification",
                                "JobQueue": "arn:aws:batch:us-east-1:123456789012:job-queue/BatchJobQueue-7049d367474b4dd",
                                "JobDefinition": "arn:aws:batch:us-east-1:123456789012:job-definition/BatchJobDefinition-74d55ec34c4643c:1",
                            },
                            "Next": "Notify Success",
                            "Catch": [
                                {
                                    "ErrorEquals": ["States.ALL"],
                                    "Next": "Notify Failure",
                                }
                            ],
                        },
                        "Notify Success": {
                            "Type": "Task",
                            "Resource": "arn:aws:states:::sns:publish",
                            "Parameters": {
                                "Message": "Batch job submitted through Step Functions succeeded",
                                "TopicArn": "arn:aws:sns:us-east-1:123456789012:batchjobnotificatiointemplate-SNSTopic-1J757CVBQ2KHM",
                            },
                            "End": True,
                        },
                        "Notify Failure": {
                            "Type": "Task",
                            "Resource": "arn:aws:states:::sns:publish",
                            "Parameters": {
                                "Message": "Batch job submitted through Step Functions failed",
                                "TopicArn": "arn:aws:sns:us-east-1:123456789012:batchjobnotificatiointemplate-SNSTopic-1J757CVBQ2KHM",
                            },
                            "End": True,
                        },
                    },
                }
            },
            [],
        ),
        (
            "Fan out example",
            {
                "Definition": {
                    "Comment": "An example of the Amazon States Language for fanning out AWS Batch job",
                    "StartAt": "Generate batch job input",
                    "TimeoutSeconds": 3600,
                    "States": {
                        "Generate batch job input": {
                            "Type": "Task",
                            "Resource": "arn:aws:states:::lambda:invoke",
                            "OutputPath": "$.Payload",
                            "Parameters": {
                                "FunctionName": "<GENERATE_BATCH_JOB_INPUT_LAMBDA_FUNCTION_NAME>"
                            },
                            "Next": "Fan out batch jobs",
                        },
                        "Fan out batch jobs": {
                            "Comment": "Start multiple executions of batch job depending on pre-processed data",
                            "Type": "Map",
                            "End": True,
                            "ItemsPath": "$",
                            "Parameters": {"BatchNumber.$": "$$.Map.Item.Value"},
                            "ItemProcessor": {
                                "StartAt": "Submit Batch Job",
                                "States": {
                                    "Submit Batch Job": {
                                        "Type": "Task",
                                        "Resource": "arn:aws:states:::batch:submitJob.sync",
                                        "Parameters": {
                                            "JobName": "BatchJobFanOut",
                                            "JobQueue": "<BATCH_QUEUE_ARN>",
                                            "JobDefinition": "<BATCH_JOB_DEFINITION_ARN>",
                                        },
                                        "End": True,
                                    }
                                },
                            },
                        },
                    },
                }
            },
            [],
        ),
        (
            "Error handling",
            {
                "Definition": """
                {
                    "Comment": "An example of the Amazon States Language for notification on an AWS Batch job completion",
                    "StartAt": "Submit Batch Job",
                    "TimeoutSeconds": 3600,
                    "States": {
                        "Submit Batch Job": {
                            "Type": "Task",
                            "Resource": "arn:aws:states:::batch:submitJob.sync",
                            "Parameters": {
                                "JobName": "BatchJobNotification",
                                "JobQueue": "arn:aws:batch:us-west-2:123456789012:job-queue/BatchJobQueue-123456789abcdef",
                                "JobDefinition": "arn:aws:batch:us-west-2:123456789012:job-definition/BatchJobDefinition-123456789abcdef:1",
                            },
                            "Next": "Notify Success",
                            "Retry": [
                                {
                                    "ErrorEquals": ["States.ALL"],
                                    "IntervalSeconds": 30,
                                    "MaxAttempts": 2,
                                    "BackoffRate": 1.5,
                                }
                            ],
                            "Catch": [
                                {"ErrorEquals": ["States.ALL"], "Next": "Notify Failure"}
                            ],
                        },
                        "Notify Success": {
                            "Type": "Task",
                            "Resource": "arn:aws:states:::sns:publish",
                            "Parameters": {
                                "Message": "Batch job submitted through Step Functions succeeded",
                                "TopicArn": "arn:aws:sns:us-west-2:123456789012:StepFunctionsSample-BatchJobManagement12345678-9abc-def0-1234-567890abcdef-SNSTopic-A2B3C4D5E6F7G",
                            },
                            "End": True,
                        },
                        "Notify Failure": {
                            "Type": "Task",
                            "Resource": "arn:aws:states:::sns:publish",
                            "Parameters": {
                                "Message": "Batch job submitted through Step Functions failed",
                                "TopicArn": "arn:aws:sns:us-west-2:123456789012:StepFunctionsSample-BatchJobManagement12345678-9abc-def0-1234-567890abcdef-SNSTopic-A2B3C4D5E6F7G",
                            },
                            "End": True,
                        }
                    }
                }
            """
            },
            [],
        ),
        (
            "Transfer data records",
            {
                "Definition": {
                    "Comment": "An example of the Amazon States Language for reading messages from a DynamoDB table and sending them to SQS",
                    "StartAt": "Seed the DynamoDB Table",
                    "TimeoutSeconds": 3600,
                    "States": {
                        "Seed the DynamoDB Table": {
                            "Type": "Task",
                            "Resource": "arn:aws:lambda:us-east-1:123456789012:function:sqsconnector-SeedingFunction-T3U43VYDU5OQ",
                            "ResultPath": "$.List",
                            "Next": "For Loop Condition",
                        },
                        "For Loop Condition": {
                            "Type": "Choice",
                            "Choices": [
                                {
                                    "Not": {
                                        "Variable": "$.List[0]",
                                        "StringEquals": "DONE",
                                    },
                                    "Next": "Read Next Message from DynamoDB",
                                }
                            ],
                            "Default": "Succeed",
                        },
                        "Read Next Message from DynamoDB": {
                            "Type": "Task",
                            "Resource": "arn:aws:states:::dynamodb:getItem",
                            "Parameters": {
                                "TableName": "sqsconnector-DDBTable-1CAFOJWP8QD6I",
                                "Key": {"MessageId": {"S.$": "$.List[0]"}},
                            },
                            "ResultPath": "$.DynamoDB",
                            "Next": "Send Message to SQS",
                        },
                        "Send Message to SQS": {
                            "Type": "Task",
                            "Resource": "arn:aws:states:::sqs:sendMessage",
                            "Parameters": {
                                "MessageBody.$": "$.DynamoDB.Item.Message.S",
                                "QueueUrl": "https://sqs.us-east-1.amazonaws.com/123456789012/sqsconnector-SQSQueue-QVGQBW134PWK",
                            },
                            "ResultPath": "$.SQS",
                            "Next": "Pop Element from List",
                        },
                        "Pop Element from List": {
                            "Type": "Pass",
                            "Parameters": {"List.$": "$.List[1:]"},
                            "Next": "For Loop Condition",
                        },
                        "Succeed": {"Type": "Succeed"},
                    },
                }
            },
            [],
        ),
        (
            "A map",
            {
                "Definition": {
                    "Comment": "An example of the Amazon States Language for reading messages from an SQS queue and iteratively processing each message.",
                    "StartAt": "Read messages from SQS Queue",
                    "States": {
                        "Read messages from SQS Queue": {
                            "Type": "Task",
                            "Resource": "arn:aws:states:::lambda:invoke",
                            "OutputPath": "$.Payload",
                            "Parameters": {
                                "FunctionName": "MapSampleProj-ReadFromSQSQueueLambda-1MY3M63RMJVA9"
                            },
                            "Next": "Are there messages to process?",
                        },
                        "Are there messages to process?": {
                            "Type": "Choice",
                            "Choices": [
                                {
                                    "Variable": "$",
                                    "StringEquals": "No messages",
                                    "Next": "Finish",
                                }
                            ],
                            "Default": "Process messages",
                        },
                        "Process messages": {
                            "Type": "Map",
                            "Next": "Finish",
                            "ItemsPath": "$",
                            "Parameters": {
                                "MessageNumber.$": "$$.Map.Item.Index",
                                "MessageDetails.$": "$$.Map.Item.Value",
                            },
                            "ItemProcessor": {
                                "StartAt": "Write message to DynamoDB",
                                "States": {
                                    "Write message to DynamoDB": {
                                        "Type": "Task",
                                        "Resource": "arn:aws:states:::dynamodb:putItem",
                                        "ResultPath": None,
                                        "Parameters": {
                                            "TableName": "MapSampleProj-DDBTable-YJDJ1MKIN6C5",
                                            "ReturnConsumedCapacity": "TOTAL",
                                            "Item": {
                                                "MessageId": {
                                                    "S.$": "$.MessageDetails.MessageId"
                                                },
                                                "Body": {
                                                    "S.$": "$.MessageDetails.Body"
                                                },
                                            },
                                        },
                                        "Next": "Remove message from SQS queue",
                                    },
                                    "Remove message from SQS queue": {
                                        "Type": "Task",
                                        "Resource": "arn:aws:states:::lambda:invoke",
                                        "InputPath": "$.MessageDetails",
                                        "ResultPath": None,
                                        "Parameters": {
                                            "FunctionName": "MapSampleProj-DeleteFromSQSQueueLambda-198J2839ZO5K2",
                                            "Payload": {
                                                "ReceiptHandle.$": "$.ReceiptHandle"
                                            },
                                        },
                                        "Next": "Publish message to SNS topic",
                                    },
                                    "Publish message to SNS topic": {
                                        "Type": "Task",
                                        "Resource": "arn:aws:states:::sns:publish",
                                        "InputPath": "$.MessageDetails",
                                        "Parameters": {
                                            "Subject": "Message from Step Functions!",
                                            "Message.$": "$.Body",
                                            "TopicArn": "arn:aws:sns:us-east-1:012345678910:MapSampleProj-SNSTopic-1CQO4HQ3IR1KN",
                                        },
                                        "End": True,
                                    },
                                },
                            },
                        },
                        "Finish": {"Type": "Succeed"},
                    },
                }
            },
            [],
        ),
        (
            "Test functions",
            {
                "Definition": {
                    "StartAt": "Task1",
                    "States": {
                        "Task1": {
                            "Type": "Task",
                            "Credentials": {
                                "Data": "input data",
                                "Algorithm": "SHA-1",
                            },
                            "Resource": "arn:aws:states:::batch:submitJob.sync",
                            "End": True,
                        },
                    },
                }
            },
            [],
        ),
        (
            "Choices",
            {
                "Definition": {
                    "StartAt": "ChoiceStateX",
                    "States": {
                        "ChoiceStateX": {
                            "Type": "Choice",
                            "Choices": [
                                {
                                    "Not": {
                                        "Variable": "$.type",
                                        "StringEquals": "Private",
                                    },
                                    "Next": "Public",
                                },
                                {
                                    "Variable": "$.value",
                                    "NumericEquals": 0,
                                    "Next": "ValueIsZero",
                                },
                                {
                                    "And": [
                                        {
                                            "Variable": "$.value",
                                            "NumericGreaterThanEquals": 20,
                                        },
                                        {"Variable": "$.value", "NumericLessThan": 30},
                                    ],
                                    "Next": "ValueInTwenties",
                                },
                                {
                                    "Variable": "$.value",
                                    "NumericLessThan": 0,
                                    "Next": "ValueIsNegative",
                                },
                            ],
                            "Default": "DefaultState",
                        },
                        "Public": {
                            "Type": "Task",
                            "Resource": "arn:aws:lambda:us-east-1:123456789012:function:Foo",
                            "Next": "NextState",
                        },
                        "ValueIsZero": {
                            "Type": "Task",
                            "Resource": "arn:aws:lambda:us-east-1:123456789012:function:Zero",
                            "Next": "NextState",
                        },
                        "ValueInTwenties": {
                            "Type": "Task",
                            "Resource": "arn:aws:lambda:us-east-1:123456789012:function:Bar",
                            "Next": "NextState",
                        },
                        "ValueIsNegative": {
                            "Type": "Succeed",
                        },
                        "DefaultState": {"Type": "Fail", "Cause": "No Matches!"},
                    },
                }
            },
            [],
        ),
        (
            "Invalid configuration",
            {
                "Definition": {
                    "StartAt": "SampleChoice",
                    "States": {
                        "NoType": {},
                        "SampleChoice": {
                            "Default": "__SucceedEntryPoint__",
                            "Type": "Choice",
                        },
                        "__SucceedEntryPoint__": {"Type": "Pass"},
                    },
                }
            },
            [
                ValidationError(
                    "'Type' is a required property",
                    rule=StateMachineDefinition(),
                    validator="required",
                    schema_path=deque(
                        [
                            "properties",
                            "States",
                            "patternProperties",
                            "^.{1,80}$",
                            "required",
                        ]
                    ),
                    path=deque(["Definition", "States", "NoType"]),
                ),
                ValidationError(
                    "'Choices' is a required property",
                    rule=StateMachineDefinition(),
                    validator="required",
                    schema_path=deque(
                        [
                            "properties",
                            "States",
                            "patternProperties",
                            "^.{1,80}$",
                            "allOf",
                            0,
                            "then",
                            "required",
                        ]
                    ),
                    path=deque(["Definition", "States", "SampleChoice"]),
                ),
                ValidationError(
                    "Only one of ['Next', 'End'] is a required property",
                    rule=StateMachineDefinition(),
                    validator="requiredXor",
                    schema_path=deque(
                        [
                            "properties",
                            "States",
                            "patternProperties",
                            "^.{1,80}$",
                            "allOf",
                            3,
                            "then",
                            "requiredXor",
                        ]
                    ),
                    path=deque(["Definition", "States", "__SucceedEntryPoint__"]),
                ),
            ],
        ),
        (
            "Invalid string definition",
            {
                "Definition": """
                    {
                        "States": {
                            "NoType": {}
                        }
                    }
                """
            },
            [
                ValidationError(
                    "'Type' is a required property at 'States/NoType'",
                    rule=StateMachineDefinition(),
                    validator="required",
                    schema_path=deque(
                        [
                            "properties",
                            "States",
                            "patternProperties",
                            "^.{1,80}$",
                            "required",
                        ]
                    ),
                    path=deque(["Definition", "States", "NoType"]),
                ),
                ValidationError(
                    "'StartAt' is a required property",
                    rule=StateMachineDefinition(),
                    validator="required",
                    schema_path=deque(
                        [
                            "required",
                        ]
                    ),
                    path=deque(["Definition"]),
                ),
            ],
        ),
        (
            "Valid string json with definition substitution",
            {
                "DefinitionSubstitutions": {"jobqueue": {"Ref": "MyJob"}},
                "Definition": {
                    "Comment": (
                        "An example of the Amazon States "
                        "Language for notification on an "
                        "AWS Batch job completion"
                    ),
                    "StartAt": "Submit Batch Job",
                    "TimeoutSeconds": 3600,
                    "States": {
                        "Submit Batch Job": {
                            "Type": "Task",
                            "Resource": "${jobqueue}",
                            "Parameters": {
                                "JobName": "BatchJobNotification",
                                "JobQueue": "arn:aws:batch:us-east-1:123456789012:job-queue/BatchJobQueue-7049d367474b4dd",
                                "JobDefinition": "arn:aws:batch:us-east-1:123456789012:job-definition/BatchJobDefinition-74d55ec34c4643c:1",
                            },
                            "Next": "Notify Success",
                            "Catch": [
                                {
                                    "ErrorEquals": ["States.ALL"],
                                    "Next": "Notify Failure",
                                }
                            ],
                        },
                        "Notify Success": {
                            "Type": "Task",
                            "Resource": "${invalid}",
                            "Parameters": {
                                "Message": "Batch job submitted through Step Functions succeeded",
                                "TopicArn": "arn:aws:sns:us-east-1:123456789012:batchjobnotificatiointemplate-SNSTopic-1J757CVBQ2KHM",
                            },
                            "End": True,
                        },
                        "Notify Failure": {
                            "Type": "Task",
                            "Resource": [],
                            "Parameters": {
                                "Message": "Batch job submitted through Step Functions failed",
                                "TopicArn": "arn:aws:sns:us-east-1:123456789012:batchjobnotificatiointemplate-SNSTopic-1J757CVBQ2KHM",
                            },
                            "End": True,
                        },
                    },
                },
            },
            [
                ValidationError(
                    (
                        "'${invalid}' does not match "
                        "'^arn:aws:([a-z]|-)+:([a-z]|[0-9]|-)*:[0-9]*:([a-z]|-)+:[a-zA-Z0-9-_.]+(:(\\\\$LATEST|[a-zA-Z0-9-_\\\\.]+))?$'"
                    ),
                    rule=StateMachineDefinition(),
                    validator="pattern",
                    schema_path=deque(
                        [
                            "properties",
                            "States",
                            "patternProperties",
                            "^.{1,80}$",
                            "allOf",
                            5,
                            "then",
                            "properties",
                            "Resource",
                            "pattern",
                        ]
                    ),
                    path=deque(["Definition", "States", "Notify Success", "Resource"]),
                ),
                ValidationError(
                    ("[] is not of type 'string'"),
                    rule=StateMachineDefinition(),
                    validator="type",
                    schema_path=deque(
                        [
                            "properties",
                            "States",
                            "patternProperties",
                            "^.{1,80}$",
                            "allOf",
                            5,
                            "then",
                            "properties",
                            "Resource",
                            "type",
                        ]
                    ),
                    path=deque(["Definition", "States", "Notify Failure", "Resource"]),
                ),
            ],
        ),
        (
            "Valid string json",
            {
                "DefinitionSubstitutions": "jobqueue",
                "Definition": {
                    "Comment": (
                        "An example of the Amazon States "
                        "Language for notification on an "
                        "AWS Batch job completion"
                    ),
                    "StartAt": "Submit Batch Job",
                    "TimeoutSeconds": 3600,
                    "States": {
                        "Submit Batch Job": {
                            "Type": "Task",
                            "Resource": "${jobqueue}",
                            "Parameters": {
                                "JobName": "BatchJobNotification",
                                "JobQueue": "arn:aws:batch:us-east-1:123456789012:job-queue/BatchJobQueue-7049d367474b4dd",
                                "JobDefinition": "arn:aws:batch:us-east-1:123456789012:job-definition/BatchJobDefinition-74d55ec34c4643c:1",
                            },
                            "Next": "Notify Success",
                            "Catch": [
                                {
                                    "ErrorEquals": ["States.ALL"],
                                    "Next": "Notify Failure",
                                }
                            ],
                        },
                        "Notify Success": {
                            "Type": "Task",
                            "Resource": "arn:aws:states:::sns:publish",
                            "Parameters": {
                                "Message": "Batch job submitted through Step Functions succeeded",
                                "TopicArn": "arn:aws:sns:us-east-1:123456789012:batchjobnotificatiointemplate-SNSTopic-1J757CVBQ2KHM",
                            },
                            "End": True,
                        },
                        "Notify Failure": {
                            "Type": "Task",
                            "Resource": "arn:aws:states:::sns:publish",
                            "Parameters": {
                                "Message": "Batch job submitted through Step Functions failed",
                                "TopicArn": "arn:aws:sns:us-east-1:123456789012:batchjobnotificatiointemplate-SNSTopic-1J757CVBQ2KHM",
                            },
                            "End": True,
                        },
                    },
                },
            },
            [
                ValidationError(
                    (
                        "'${jobqueue}' does not match "
                        "'^arn:aws:([a-z]|-)+:([a-z]|[0-9]|-)*:[0-9]*:([a-z]|-)+:[a-zA-Z0-9-_.]+(:(\\\\$LATEST|[a-zA-Z0-9-_\\\\.]+))?$'"
                    ),
                    rule=StateMachineDefinition(),
                    validator="pattern",
                    schema_path=deque(
                        [
                            "properties",
                            "States",
                            "patternProperties",
                            "^.{1,80}$",
                            "allOf",
                            5,
                            "then",
                            "properties",
                            "Resource",
                            "pattern",
                        ]
                    ),
                    path=deque(
                        ["Definition", "States", "Submit Batch Job", "Resource"]
                    ),
                ),
            ],
        ),
        (
            "Using a function",
            {"Definition": {"Fn::Join": ["\n", []]}},
            [],
        ),
        (
            "Invalid query languages",
            {
                "DefinitionSubstitutions": {"jobqueue": {"Ref": "MyJob"}},
                "Definition": {
                    "Comment": (
                        "An example of the Amazon States "
                        "Language for notification on an "
                        "AWS Batch job completion"
                    ),
                    "StartAt": "Submit Batch Job",
                    "TimeoutSeconds": 3600,
                    "States": {
                        "Submit Batch Job 1": {
                            "Type": "Task",
                            "Resource": "${jobqueue}",
                            "Parameters": {  # fails because JSONata doesn't support Parameters
                                "JobName": "BatchJobNotification",
                                "JobQueue": "arn:aws:batch:us-east-1:123456789012:job-queue/BatchJobQueue-7049d367474b4dd",
                                "JobDefinition": "arn:aws:batch:us-east-1:123456789012:job-definition/BatchJobDefinition-74d55ec34c4643c:1",
                            },
                            "Next": "Notify Success",
                            "Catch": [
                                {
                                    "ErrorEquals": ["States.ALL"],
                                    "Next": "Notify Failure",
                                }
                            ],
                            "QueryLanguage": "JSONata",
                        },
                        "Submit Batch Job 2": {
                            "Type": "Task",
                            "Resource": "${jobqueue}",
                            "Arguments": {  # fails because JSONPath doesn't support Arguments
                                "JobName": "BatchJobNotification",
                                "JobQueue": "arn:aws:batch:us-east-1:123456789012:job-queue/BatchJobQueue-7049d367474b4dd",
                                "JobDefinition": "arn:aws:batch:us-east-1:123456789012:job-definition/BatchJobDefinition-74d55ec34c4643c:1",
                            },
                            "Next": "Notify Success",
                            "Catch": [
                                {
                                    "ErrorEquals": ["States.ALL"],
                                    "Next": "Notify Failure",
                                }
                            ],
                            "QueryLanguage": "JSONPath",
                        },
                        "Submit Batch Job 3": {
                            "Type": "Task",
                            "Resource": "${jobqueue}",
                            "Arguments": {  # fails because JSONPath is the default for Query Language
                                "JobName": "BatchJobNotification",
                                "JobQueue": "arn:aws:batch:us-east-1:123456789012:job-queue/BatchJobQueue-7049d367474b4dd",
                                "JobDefinition": "arn:aws:batch:us-east-1:123456789012:job-definition/BatchJobDefinition-74d55ec34c4643c:1",
                            },
                            "Next": "Notify Success",
                            "Catch": [
                                {
                                    "ErrorEquals": ["States.ALL"],
                                    "Next": "Notify Failure",
                                }
                            ],
                        },
                        "Choices1": {
                            "Type": "Choice",
                            "Choices": [
                                {
                                    "Condition": "BatchJobNotification",
                                    "Next": "Notify Success",
                                }
                            ],
                            "Next": "Notify Success",
                            "QueryLanguage": "JSONata",
                        },
                        "Choices2": {
                            "Type": "Choice",
                            "Choices": [
                                {
                                    "Condition": "BatchJobNotification",
                                    "Next": "Notify Success",
                                },
                                {
                                    "Variable": "$.value",
                                    "NumericLessThan": 0,
                                    # missing next
                                },
                            ],
                            "Next": "Notify Success",
                        },
                        "Choices3": {
                            "Type": "Choice",
                            "Choices": [
                                {
                                    "And": "BatchJobNotification",
                                    "Next": "Notify Success",
                                }
                            ],
                            "Next": "Notify Success",
                            "QueryLanguage": "JSONata",
                        },
                        "Notify Success": {
                            "Type": "Task",
                            "Resource": "arn:aws:states:::sns:publish",
                            "Parameters": {
                                "Message": "Batch job submitted through Step Functions succeeded",
                                "TopicArn": "arn:aws:sns:us-east-1:123456789012:batchjobnotificatiointemplate-SNSTopic-1J757CVBQ2KHM",
                            },
                            "End": True,
                        },
                        "Notify Failure": {
                            "Type": "Task",
                            "Resource": "arn:aws:states:::sns:publish",
                            "Parameters": {
                                "Message": "Batch job submitted through Step Functions failed",
                                "TopicArn": "arn:aws:sns:us-east-1:123456789012:batchjobnotificatiointemplate-SNSTopic-1J757CVBQ2KHM",
                            },
                            "End": True,
                        },
                    },
                },
            },
            [
                ValidationError(
                    (
                        "Additional properties are not allowed ('Parameters' was unexpected)"
                    ),
                    rule=StateMachineDefinition(),
                    validator=None,
                    schema_path=deque(
                        [
                            "properties",
                            "States",
                            "patternProperties",
                            "^.{1,80}$",
                            "allOf",
                            5,
                            "then",
                            "allOf",
                            0,
                            "then",
                            "properties",
                            "Parameters",
                        ]
                    ),
                    path=deque(
                        ["Definition", "States", "Submit Batch Job 1", "Parameters"]
                    ),
                ),
                ValidationError(
                    (
                        "Additional properties are not allowed ('Arguments' was unexpected)"
                    ),
                    rule=StateMachineDefinition(),
                    validator=None,
                    schema_path=deque(
                        [
                            "properties",
                            "States",
                            "patternProperties",
                            "^.{1,80}$",
                            "allOf",
                            5,
                            "then",
                            "allOf",
                            1,
                            "then",
                            "properties",
                            "Arguments",
                        ]
                    ),
                    path=deque(
                        ["Definition", "States", "Submit Batch Job 2", "Arguments"]
                    ),
                ),
                ValidationError(
                    (
                        "Additional properties are not allowed ('Arguments' was unexpected)"
                    ),
                    rule=StateMachineDefinition(),
                    validator=None,
                    schema_path=deque(
                        [
                            "properties",
                            "States",
                            "patternProperties",
                            "^.{1,80}$",
                            "allOf",
                            5,
                            "then",
                            "allOf",
                            1,
                            "then",
                            "properties",
                            "Arguments",
                        ]
                    ),
                    path=deque(
                        ["Definition", "States", "Submit Batch Job 3", "Arguments"]
                    ),
                ),
                ValidationError(
                    (
                        "Only one of ['And', 'BooleanEquals', 'BooleanEqualsPath', 'IsBoolean', 'IsNull', "
                        "'IsNumeric', 'IsPresent', 'IsString', 'IsTimestamp', 'Not', 'NumericEquals', "
                        "'NumericEqualsPath', 'NumericGreaterThan', 'NumericGreaterThanPath', "
                        "'NumericGreaterThanEquals', 'NumericGreaterThanEqualsPath', 'NumericLessThan', "
                        "'NumericLessThanPath', 'NumericLessThanEquals', 'NumericLessThanEqualsPath', 'Or', "
                        "'StringEquals', 'StringEqualsPath', 'StringGreaterThan', 'StringGreaterThanPath', "
                        "'StringGreaterThanEquals', 'StringGreaterThanEqualsPath', 'StringLessThan', "
                        "'StringLessThanPath', 'StringLessThanEquals', 'StringLessThanEqualsPath', "
                        "'StringMatches', 'TimestampEquals', 'TimestampEqualsPath', 'TimestampGreaterThan', "
                        "'TimestampGreaterThanPath', 'TimestampGreaterThanEquals', 'TimestampGreaterThanEqualsPath', "
                        "'TimestampLessThan', 'TimestampLessThanPath', 'TimestampLessThanEquals', "
                        "'TimestampLessThanEqualsPath'] is a required property"
                    ),
                    rule=StateMachineDefinition(),
                    validator="requiredXor",
                    schema_path=deque(
                        [
                            "properties",
                            "States",
                            "patternProperties",
                            "^.{1,80}$",
                            "allOf",
                            0,
                            "then",
                            "allOf",
                            1,
                            "then",
                            "properties",
                            "Choices",
                            "items",
                            "requiredXor",
                        ]
                    ),
                    path=deque(["Definition", "States", "Choices2", "Choices", 0]),
                ),
                ValidationError(
                    ("'Next' is a required property"),
                    rule=StateMachineDefinition(),
                    validator="required",
                    schema_path=deque(
                        [
                            "properties",
                            "States",
                            "patternProperties",
                            "^.{1,80}$",
                            "allOf",
                            0,
                            "then",
                            "allOf",
                            1,
                            "then",
                            "properties",
                            "Choices",
                            "items",
                            "required",
                        ]
                    ),
                    path=deque(["Definition", "States", "Choices2", "Choices", 1]),
                ),
                ValidationError(
                    ("'Condition' is a required property"),
                    rule=StateMachineDefinition(),
                    validator="required",
                    schema_path=deque(
                        [
                            "properties",
                            "States",
                            "patternProperties",
                            "^.{1,80}$",
                            "allOf",
                            0,
                            "then",
                            "allOf",
                            0,
                            "then",
                            "properties",
                            "Choices",
                            "items",
                            "required",
                        ]
                    ),
                    path=deque(["Definition", "States", "Choices3", "Choices", 0]),
                ),
            ],
        ),
        (
            "JSONAta set at the top level",
            {
                "DefinitionSubstitutions": {"jobqueue": {"Ref": "MyJob"}},
                "Definition": {
                    "Comment": (
                        "An example of the Amazon States "
                        "Language for notification on an "
                        "AWS Batch job completion"
                    ),
                    "StartAt": "Submit Batch Job",
                    "TimeoutSeconds": 3600,
                    "QueryLanguage": "JSONata",
                    "States": {
                        "Submit Batch Job 1": {
                            "Type": "Task",
                            "Resource": "${jobqueue}",
                            "QueryLanguage": "JSONPath",  # fails because QueryLanguage for JSONata can't be over written
                            "Parameters": {  # fails because JSONata doesn't support Parameters
                                "JobName": "BatchJobNotification",
                                "JobQueue": "arn:aws:batch:us-east-1:123456789012:job-queue/BatchJobQueue-7049d367474b4dd",
                                "JobDefinition": "arn:aws:batch:us-east-1:123456789012:job-definition/BatchJobDefinition-74d55ec34c4643c:1",
                            },
                            "Next": "Notify Success",
                            "Catch": [
                                {
                                    "ErrorEquals": ["States.ALL"],
                                    "Next": "Notify Failure",
                                }
                            ],
                        },
                        "Submit Batch Job 2": {
                            "Type": "Task",
                            "Resource": "${jobqueue}",
                            "Arguments": {
                                "JobName": "BatchJobNotification",
                                "JobQueue": "arn:aws:batch:us-east-1:123456789012:job-queue/BatchJobQueue-7049d367474b4dd",
                                "JobDefinition": "arn:aws:batch:us-east-1:123456789012:job-definition/BatchJobDefinition-74d55ec34c4643c:1",
                            },
                            "Next": "Notify Success",
                            "Catch": [
                                {
                                    "ErrorEquals": ["States.ALL"],
                                    "Next": "Notify Failure",
                                }
                            ],
                            "QueryLanguage": "JSONata",
                        },
                        "Submit Batch Job 3": {
                            "Type": "Task",
                            "Resource": "${jobqueue}",
                            "Arguments": {
                                "JobName": "BatchJobNotification",
                                "JobQueue": "arn:aws:batch:us-east-1:123456789012:job-queue/BatchJobQueue-7049d367474b4dd",
                                "JobDefinition": "arn:aws:batch:us-east-1:123456789012:job-definition/BatchJobDefinition-74d55ec34c4643c:1",
                            },
                            "Next": "Notify Success",
                            "Catch": [
                                {
                                    "ErrorEquals": ["States.ALL"],
                                    "Next": "Notify Failure",
                                }
                            ],
                        },
                        "Notify Success": {
                            "Type": "Task",
                            "Resource": "arn:aws:states:::sns:publish",
                            "Parameters": {  # fails because the default is now JSONata
                                "Message": "Batch job submitted through Step Functions succeeded",
                                "TopicArn": "arn:aws:sns:us-east-1:123456789012:batchjobnotificatiointemplate-SNSTopic-1J757CVBQ2KHM",
                            },
                            "End": True,
                        },
                        "Choices1": {
                            "Type": "Choice",
                            "Choices": [
                                {
                                    "Condition": "{% $states.input.type != 'Private' %}",
                                    "Next": "Notify Failure",
                                },
                            ],
                            "End": True,
                        },
                        "Choices2": {
                            "Type": "Choice",
                            "Choices": [
                                {
                                    "And": "{% $states.input.type != 'Private' %}",
                                    "Next": "Notify Failure",
                                },
                            ],
                            "End": True,
                        },
                        "Notify Failure": {
                            "Type": "Task",
                            "Resource": "arn:aws:states:::sns:publish",
                            "Parameters": {  # fails because the default is now JSONata
                                "Message": "Batch job submitted through Step Functions failed",
                                "TopicArn": "arn:aws:sns:us-east-1:123456789012:batchjobnotificatiointemplate-SNSTopic-1J757CVBQ2KHM",
                            },
                            "End": True,
                        },
                    },
                },
            },
            [
                ValidationError(
                    ("'JSONata' was expected"),
                    rule=StateMachineDefinition(),
                    validator="const",
                    schema_path=deque(
                        [
                            "properties",
                            "States",
                            "patternProperties",
                            "^.{1,80}$",
                            "allOf",
                            5,
                            "then",
                            "allOf",
                            0,
                            "properties",
                            "QueryLanguage",
                            "const",
                        ]
                    ),
                    path=deque(
                        ["Definition", "States", "Submit Batch Job 1", "QueryLanguage"]
                    ),
                ),
                ValidationError(
                    (
                        "Additional properties are not allowed ('Parameters' was unexpected)"
                    ),
                    rule=StateMachineDefinition(),
                    validator=None,
                    schema_path=deque(
                        [
                            "properties",
                            "States",
                            "patternProperties",
                            "^.{1,80}$",
                            "allOf",
                            5,
                            "then",
                            "allOf",
                            0,
                            "properties",
                            "Parameters",
                        ]
                    ),
                    path=deque(
                        ["Definition", "States", "Submit Batch Job 1", "Parameters"]
                    ),
                ),
                ValidationError(
                    (
                        "Additional properties are not allowed ('Parameters' was unexpected)"
                    ),
                    rule=StateMachineDefinition(),
                    validator=None,
                    schema_path=deque(
                        [
                            "properties",
                            "States",
                            "patternProperties",
                            "^.{1,80}$",
                            "allOf",
                            5,
                            "then",
                            "allOf",
                            0,
                            "properties",
                            "Parameters",
                        ]
                    ),
                    path=deque(
                        ["Definition", "States", "Notify Success", "Parameters"]
                    ),
                ),
                ValidationError(
                    ("'Condition' is a required property"),
                    rule=StateMachineDefinition(),
                    validator="required",
                    schema_path=deque(
                        [
                            "properties",
                            "States",
                            "patternProperties",
                            "^.{1,80}$",
                            "allOf",
                            0,
                            "then",
                            "allOf",
                            0,
                            "properties",
                            "Choices",
                            "items",
                            "required",
                        ]
                    ),
                    path=deque(["Definition", "States", "Choices2", "Choices", 0]),
                ),
                ValidationError(
                    ("Additional properties are not allowed ('And' was unexpected)"),
                    rule=StateMachineDefinition(),
                    validator="additionalProperties",
                    schema_path=deque(
                        [
                            "properties",
                            "States",
                            "patternProperties",
                            "^.{1,80}$",
                            "allOf",
                            0,
                            "then",
                            "allOf",
                            0,
                            "properties",
                            "Choices",
                            "items",
                            "additionalProperties",
                        ]
                    ),
                    path=deque(
                        ["Definition", "States", "Choices2", "Choices", 0, "And"]
                    ),
                ),
                ValidationError(
                    (
                        "Additional properties are not allowed ('Parameters' was unexpected)"
                    ),
                    rule=StateMachineDefinition(),
                    validator=None,
                    schema_path=deque(
                        [
                            "properties",
                            "States",
                            "patternProperties",
                            "^.{1,80}$",
                            "allOf",
                            5,
                            "then",
                            "allOf",
                            0,
                            "properties",
                            "Parameters",
                        ]
                    ),
                    path=deque(
                        ["Definition", "States", "Notify Failure", "Parameters"]
                    ),
                ),
            ],
        ),
    ],
)
def test_validate(
    name,
    instance,
    expected,
    rule,
    validator,
):
    errs = list(rule.validate(validator, {}, instance, {}))
    assert errs == expected, f"{name!r} test failed with {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from test.testlib.testcase import BaseTestCase
from typing import Any

from cfnlint.rules import CloudFormationLintRule, Match, RuleMatch
from cfnlint.rules._rule import matching


class rule(CloudFormationLintRule):
    id: str = "ETest"
    pass


rule_match = RuleMatch([], "Message", rule=rule(), location=(0, 0, 0, 0))


class TestMatching(BaseTestCase):
    """Test Matching Wrapper"""

    def example(self: Any, *args: Any, **kwargs: Any):
        return [rule_match]

    def test_matching_location(self):
        f = matching("example")
        r = f(self.example)
        t = r(self, "", None)
        self.assertEqual(
            list(t),
            [
                Match(
                    linenumber=1,
                    columnnumber=1,
                    linenumberend=1,
                    columnnumberend=1,
                    filename="",
                    rule=rule(),
                    message="Message",
                    rulematch_obj=rule_match,
                )
            ],
        )

    def test_compare(self):
        rule_match_1 = RuleMatch(
            ["path"], "Message", rule=rule(), location=(0, 0, 0, 0)
        )
        rule_match_2 = RuleMatch(
            ["path"], "Message", rule=rule(), location=(0, 0, 0, 0)
        )
        self.assertEqual(rule_match_1, rule_match_2)

        rule_match_3 = RuleMatch(
            ["path"], "New Message", rule=rule(), location=(1, 1, 1, 1)
        )
        self.assertNotEqual(rule_match_1, rule_match_3)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import logging

import pytest

from cfnlint.schema._schema import Schema

LOGGER = logging.getLogger("cfnlint.schema.manager")
LOGGER.disabled = True


@pytest.mark.parametrize(
    "name,resource_schema,expected",
    [
        (
            "Standard ref",
            {
                "additionalProperties": False,
                "createOnlyProperties": [],
                "primaryIdentifier": ["/properties/One"],
                "properties": {"One": {"type": "string"}, "Two": {"type": "boolean"}},
                "readOnlyProperties": [],
                "typeName": "Foo::Foo::Foo",
                "writeOnlyProperties": [],
            },
            {"type": "string"},
        ),
        (
            "Standard with multiple identifiers",
            {
                "additionalProperties": False,
                "createOnlyProperties": [],
                "primaryIdentifier": ["/properties/One", "/properties/Two"],
                "properties": {"One": {"type": "integer"}, "Two": {"type": "boolean"}},
                "readOnlyProperties": [],
                "typeName": "Foo::Foo::Foo",
                "writeOnlyProperties": [],
            },
            {"type": "string"},
        ),
    ],
)
def test_schema(name, resource_schema, expected):

    schema = Schema(resource_schema)

    assert schema.ref == expected, f"{name!r} test got {schema.ref!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.ectwo.SecurityGroupProtocolsAndPortsInclusive import (
    SecurityGroupProtocolsAndPortsInclusive,
)


@pytest.fixture(scope="module")
def rule():
    rule = SecurityGroupProtocolsAndPortsInclusive()
    yield rule


@pytest.mark.parametrize(
    "instance,expected",
    [
        (
            {
                "IpProtocol": -1,
            },
            [],
        ),
        (
            {
                "IpProtocol": 9,
            },
            [],
        ),
        (
            [],  # wrong type
            [],
        ),
        (
            {
                "IpProtocol": "tcp",
                "FromPort": 1,
            },
            [
                ValidationError(
                    (
                        "['FromPort', 'ToPort'] are required properties "
                        "when using 'IpProtocol' value 'tcp'"
                    ),
                    rule=SecurityGroupProtocolsAndPortsInclusive(),
                    path=deque([]),
                    instance={"IpProtocol": "tcp", "FromPort": 1},
                    validator="required",
                    validator_value=["FromPort", "ToPort"],
                    schema_path=deque(["then", "required"]),
                )
            ],
        ),
    ],
)
def test_backup_lifecycle(instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))

    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright (c) 2013 Julian Berman

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

SPDX-License-Identifier: MIT
"""

# Code is taken from jsonschema package and adapted CloudFormation use
# https://github.com/python-jsonschema/jsonschema
from __future__ import annotations

import datetime
import ipaddress
import typing
from dataclasses import dataclass, field

import regex as re

from cfnlint.jsonschema.exceptions import FormatError

_FormatCheckCallable = typing.Callable[[object], bool]
_RaisesType = typing.Union[
    typing.Type[Exception],
    typing.Tuple[typing.Type[Exception], ...],
]

_RE_DATE_FULLYEAR = "\\d{4}"
_RE_DATE_MONTH = "(0[1-9]|1[0-2])"
_RE_DATE_MDAY = "\\d{2}"
_RE_TIME_HOUR = "([01]\\d|2[0123])"
_RE_TIME_MINUTE = "[0-5]\\d"
_RE_TIME_SECOND = "[0-5]\\d"
_RE_TIME_SECRFAC = "(\\.\\d+)?"
_RE_TIME_NUMOFFSET = f"[+-]{_RE_TIME_HOUR}:{_RE_TIME_MINUTE}"
_RE_TIME_OFFSET = f"(Z|[+-]{_RE_TIME_NUMOFFSET})"
_RE_PARTIAL_TIME = (
    f"{_RE_TIME_HOUR}:{_RE_TIME_MINUTE}:{_RE_TIME_SECOND}{_RE_TIME_SECRFAC}"
)
_RE_FULL_DATE = f"{_RE_DATE_FULLYEAR}-{_RE_DATE_MONTH}-{_RE_DATE_MDAY}"
_RE_FULL_TIME = f"{_RE_PARTIAL_TIME}{_RE_TIME_OFFSET}"

_RE_DATE = re.compile(
    rf"^{_RE_FULL_DATE}$",
    re.ASCII,
)
_RE_DATE_TIME = re.compile(
    rf"^{_RE_FULL_DATE}T{_RE_FULL_TIME}$",
    re.ASCII,
)


@dataclass
class FormatChecker:
    """
    A ``format`` property checker.

    JSON Schema does not mandate that the ``format`` property actually do any
    validation. If validation is desired however, instances of this class can
    be hooked into validators to enable format validation.

    `FormatChecker` objects always return ``True`` when asked about
    formats that they do not know how to validate.

    To add a check for a custom format use the `FormatChecker.checks`
    decorator.

    Arguments:

        formats:

            The known formats to validate. This argument can be used to
            limit which formats will be used during validation.
    """

    checkers: dict[
        str,
        tuple[_FormatCheckCallable, _RaisesType],
    ] = field(default_factory=dict)

    def __repr__(self):
        return f"<FormatChecker checkers={sorted(self.checkers)}>"

    def check(self, instance: object, format: str) -> None:
        """
        Check whether the instance conforms to the given format.

        Arguments:

            instance (*any primitive type*, i.e. str, number, bool):

                The instance to check

            format:

                The format that instance should conform to

        Raises:

            FormatError:

                if the instance does not conform to ``format``
        """
        if format not in self.checkers:
            return

        func, raises = self.checkers[format]
        result, cause = None, None
        try:
            result = func(instance)
        except raises as e:
            cause = e
        if not result:
            raise FormatError(f"{instance!r} is not a {format!r}", cause=cause)


def is_email(instance: object) -> bool:
    if not isinstance(instance, str):
        return True
    return "@" in instance


def is_ipv4(instance: object) -> bool:
    if not isinstance(instance, str):
        return True
    try:
        return bool(ipaddress.IPv4Address(instance))
    except ipaddress.AddressValueError:
        return False


def is_ipv4_network(instance: object) -> bool:
    if not isinstance(instance, str):
        return True
    try:
        return bool(ipaddress.IPv4Network(instance))
    except (ipaddress.AddressValueError, ValueError):
        return False


def is_ipv6(instance: object) -> bool:
    if not isinstance(instance, str):
        return True
    try:
        address = ipaddress.IPv6Address(instance)
        return not getattr(address, "scope_id", "")
    except ipaddress.AddressValueError:
        return False


def is_ipv6_network(instance: object) -> bool:
    if not isinstance(instance, str):
        return True
    try:
        return bool(ipaddress.IPv6Network(instance))
    except (ipaddress.AddressValueError, ValueError):
        return False


def is_datetime(instance: object) -> bool:
    if not isinstance(instance, str):
        return True
    return bool(_RE_DATE_TIME.fullmatch(instance))


def is_time(instance: object) -> bool:
    if not isinstance(instance, str):
        return True
    return bool(_RE_DATE_TIME.fullmatch(f"1970-01-01T{instance}"))


def is_regex(instance: object) -> bool:
    if not isinstance(instance, str):
        return True
    return bool(re.compile(instance))


def is_date(instance: object) -> bool:
    if not isinstance(instance, str):
        return True
    return bool(_RE_DATE.fullmatch(instance) and datetime.date.fromisoformat(instance))


cfn_format_checker = FormatChecker(
    {
        "date": (is_date, ()),
        "date-time": (is_datetime, ()),
        "email": (is_email, ()),
        "ipv4": (is_ipv4, ipaddress.AddressValueError),
        "ipv4-network": (is_ipv4_network, ipaddress.AddressValueError),
        "ipv6": (is_ipv6, ipaddress.AddressValueError),
        "ipv6-network": (is_ipv6_network, ipaddress.AddressValueError),
        "regex": (is_regex, re.error),
        "time": (is_time, ()),
    }
)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from typing import Any

import cfnlint.data.AdditionalSpecs
from cfnlint.helpers import load_resource
from cfnlint.jsonschema import Validator
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema


class UpdateReplacePolicyDeletionPolicyOnStatefulResourceTypes(CfnLintJsonSchema):
    id = "I3011"
    shortdesc = "Check stateful resources have a set UpdateReplacePolicy/DeletionPolicy"
    description = (
        "The default action when replacing/removing a resource is to "
        "delete it. This check requires you to explicitly set policies"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-attribute-deletionpolicy.html"
    tags = ["resources", "updatereplacepolicy", "deletionpolicy"]

    def __init__(self):
        """Init"""
        super().__init__(
            keywords=["Resources/*"],
            all_matches=True,
        )

        spec = load_resource(cfnlint.data.AdditionalSpecs, "StatefulResources.json")
        self.likely_stateful_resource_types = [
            resource_type
            for resource_type, descr in spec["ResourceTypes"].items()
            # Resources that won't be deleted if they're not empty (ex: S3)
            # don't need to be checked for policies, as chance of mistakes are low.
            if not descr.get("DeleteRequiresEmptyResource", False)
        ]

        self._schema = {"required": ["DeletionPolicy", "UpdateReplacePolicy"]}

    def validate(self, validator: Validator, s: Any, instance: Any, schema: Any):
        resource_type = instance.get("Type")

        if not isinstance(resource_type, str):
            return

        if resource_type not in self.likely_stateful_resource_types:  # type: ignore
            return

        for err in super().validate(validator, s, instance, self._schema):
            err.message = (
                f"{err.message} (The default action when replacing/removing "
                "a resource is to delete it. Set explicit values for "
                "stateful resource)"
            )
            yield err



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import os

from cfnlint._typing import RuleMatches
from cfnlint.decode import decode
from cfnlint.rules import CloudFormationLintRule, RuleMatch
from cfnlint.template.template import Template


class NestedStackParameters(CloudFormationLintRule):
    """Check that nested stack parameters are specified as needed"""

    id = "E3043"
    shortdesc = "Validate parameters for in a nested stack"
    description = (
        "Evalute if parameters for a nested stack are specified and "
        "if parameters are specified for a nested stack that aren't required."
    )
    source_url = "https://github.com/aws-cloudformation/cfn-lint"
    tags = ["resources", "cloudformation"]

    def __init__(self):
        """Init"""
        super().__init__()
        self.resource_property_types.append("AWS::CloudFormation::Stack")

    def __get_template_parameters(self, filename):
        try:
            (tmp, matches) = decode(filename)
        except:  # noqa: E722
            return None
        if matches:
            return None

        return tmp.get("Parameters", {})

    def __compare_objects(self, template_parameters, nested_parameters, scenario, path):
        matches = []
        for key in set(template_parameters.keys()) - set(nested_parameters.keys()):
            if scenario is None:
                message = (
                    'Specified parameter "{0}" doesn\'t exist in nested stack template'
                    " at {1}"
                )
                matches.append(
                    RuleMatch(
                        path + [key],
                        message.format(key, "/".join(map(str, path + [key]))),
                    )
                )
            else:
                message = (
                    'Specified parameter "{0}" doesn\'t exist in nested stack'
                    " template {1}"
                )
                scenario_text = " and ".join(
                    [f'when condition "{k}" is {v}' for (k, v) in scenario.items()]
                )
                matches.append(RuleMatch(path, message.format(key, scenario_text)))
        for key in set(nested_parameters.keys()) - set(template_parameters.keys()):
            if nested_parameters.get(key).get("Default") is None:
                if scenario is None:
                    message = (
                        'Nested stack template parameter "{0}" is not specified at {1}'
                    )
                    matches.append(
                        RuleMatch(path, message.format(key, "/".join(map(str, path))))
                    )
                else:
                    message = (
                        'Nested stack template parameter "{0}" is not specified {1}'
                    )
                    scenario_text = " and ".join(
                        [f'when condition "{k}" is {v}' for (k, v) in scenario.items()]
                    )
                    matches.append(RuleMatch(path, message.format(key, scenario_text)))

        return matches

    def match(self, cfn: Template) -> RuleMatches:
        """Check CloudFormation Properties"""
        matches = []
        resources = cfn.get_resources("AWS::CloudFormation::Stack")
        for resource_name, attributes in resources.items():
            properties = attributes.get("Properties", {})
            # when template is passed via cat (or equivalent filename is none)
            if cfn.filename:
                base_dir = os.path.dirname(os.path.abspath(cfn.filename))

                parameter_groups = cfn.get_object_without_conditions(
                    obj=properties, property_names=["TemplateURL", "Parameters"]
                )

                for parameter_group in parameter_groups:
                    obj = parameter_group.get("Object")
                    template_url = obj.get("TemplateURL")
                    if isinstance(template_url, str):
                        if not (
                            template_url.startswith("http://")
                            or template_url.startswith("https://")
                            or template_url.startswith("s3://")
                        ):
                            template_path = os.path.normpath(
                                os.path.join(base_dir, template_url)
                            )
                            nested_parameters = self.__get_template_parameters(
                                template_path
                            )
                            template_parameters = obj.get("Parameters")
                            if isinstance(nested_parameters, dict) and isinstance(
                                template_parameters, dict
                            ):
                                matches.extend(
                                    self.__compare_objects(
                                        template_parameters=template_parameters,
                                        nested_parameters=nested_parameters,
                                        path=[
                                            "Resources",
                                            resource_name,
                                            "Properties",
                                            "Parameters",
                                        ],
                                        scenario=parameter_group.get("Scenario"),
                                    )
                                )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.cloudfront.Aliases import Aliases


@pytest.fixture(scope="module")
def rule():
    rule = Aliases()
    yield rule


@pytest.mark.parametrize(
    "instance,expected",
    [
        (
            "www.example.com",
            [],
        ),
        (
            "example.com",
            [],
        ),
        (
            "email.exa.ple.com",
            [],
        ),
        (
            "mx1.example.eu",
            [],
        ),
        (
            "e-mail.example.amsterdam",
            [],
        ),
        ("xn-caf-dma.com", []),
        ("www.example.eu", []),
        (
            "111.example.com",
            [],
        ),
        (
            "email.internal.example.eu",
            [],
        ),
        (
            "e-mail.internal.ex-ample.eu",
            [],
        ),
        (
            "www.example.google",
            [],
        ),
        (
            "1111111111.ex--ample.nl",
            [],
        ),
        (
            "*.example.com",
            [],
        ),
        (
            {},
            [],
        ),
        (
            "email.*.example.com",
            [
                ValidationError(
                    (
                        "'email.*.example.com' does not match "
                        "'^(?!.*(?:\\\\.\\\\*\\\\.)).*'"
                    ),
                    rule=Aliases(),
                    path=deque([]),
                    validator="pattern",
                    schema_path=deque(["pattern"]),
                )
            ],
        ),
        (
            "WWW.EXAMPLE.COM",
            [
                ValidationError(
                    (
                        "'WWW.EXAMPLE.COM' does not match "
                        "'^(?:[a-z0-9\\\\*](?:[a-z0-9-]{0,61}[a-z0-9])?\\\\.)+[a-z0-9][a-z0-9-]{0,61}[a-z0-9]$'"
                    ),
                    rule=Aliases(),
                    path=deque([]),
                    validator="pattern",
                    schema_path=deque(["pattern"]),
                )
            ],
        ),
        (
            "-example.com",
            [
                ValidationError(
                    (
                        "'-example.com' does not match "
                        "'^(?:[a-z0-9\\\\*](?:[a-z0-9-]{0,61}[a-z0-9])?\\\\.)+[a-z0-9][a-z0-9-]{0,61}[a-z0-9]$'"
                    ),
                    rule=Aliases(),
                    path=deque([]),
                    validator="pattern",
                    schema_path=deque(["pattern"]),
                )
            ],
        ),
        (
            "example.c",
            [
                ValidationError(
                    (
                        "'example.c' does not match "
                        "'^(?:[a-z0-9\\\\*](?:[a-z0-9-]{0,61}[a-z0-9])?\\\\.)+[a-z0-9][a-z0-9-]{0,61}[a-z0-9]$'"
                    ),
                    rule=Aliases(),
                    path=deque([]),
                    validator="pattern",
                    schema_path=deque(["pattern"]),
                )
            ],
        ),
        (
            "www.example.com ",
            [
                ValidationError(
                    (
                        "'www.example.com ' does not match "
                        "'^(?:[a-z0-9\\\\*](?:[a-z0-9-]{0,61}[a-z0-9])?\\\\.)+[a-z0-9][a-z0-9-]{0,61}[a-z0-9]$'"
                    ),
                    rule=Aliases(),
                    path=deque([]),
                    validator="pattern",
                    schema_path=deque(["pattern"]),
                )
            ],
        ),
        (
            "www.-example.com",
            [
                ValidationError(
                    (
                        "'www.-example.com' does not match "
                        "'^(?:[a-z0-9\\\\*](?:[a-z0-9-]{0,61}[a-z0-9])?\\\\.)+[a-z0-9][a-z0-9-]{0,61}[a-z0-9]$'"
                    ),
                    rule=Aliases(),
                    path=deque([]),
                    validator="pattern",
                    schema_path=deque(["pattern"]),
                )
            ],
        ),
        (
            "-www.example.com",
            [
                ValidationError(
                    (
                        "'-www.example.com' does not match "
                        "'^(?:[a-z0-9\\\\*](?:[a-z0-9-]{0,61}[a-z0-9])?\\\\.)+[a-z0-9][a-z0-9-]{0,61}[a-z0-9]$'"
                    ),
                    rule=Aliases(),
                    path=deque([]),
                    validator="pattern",
                    schema_path=deque(["pattern"]),
                )
            ],
        ),
        (
            "www.example.com-",
            [
                ValidationError(
                    (
                        "'www.example.com-' does not match "
                        "'^(?:[a-z0-9\\\\*](?:[a-z0-9-]{0,61}[a-z0-9])?\\\\.)+[a-z0-9][a-z0-9-]{0,61}[a-z0-9]$'"
                    ),
                    rule=Aliases(),
                    path=deque([]),
                    validator="pattern",
                    schema_path=deque(["pattern"]),
                )
            ],
        ),
    ],
)
def test_backup_lifecycle(instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))

    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from test.unit.rules import BaseRuleTestCase

from cfnlint.rules.custom.Operators import CreateGreaterRule  # pylint: disable=E0401


class TestGreaterOperator(BaseRuleTestCase):
    """Test template mapping configurations"""

    def setUp(self):
        """Setup"""
        super(TestGreaterOperator, self).setUp()
        self.collection.register(
            CreateGreaterRule(
                "E9001",
                "AWS::Lambda::Function",
                "Timeout",
                "50",
                "Timeout should be greater than 50",
            )
        )

    success_templates = [
        "test/fixtures/templates/good/custom/numeric-inequalities-large.yaml"
    ]

    def test_file_positive(self):
        """Test Positive"""
        self.helper_file_positive()

    def test_file_negative(self):
        """Test failure"""
        self.helper_file_negative(
            "test/fixtures/templates/good/custom/numeric-inequalities-small.yaml", 2
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.jsonschema._keywords import enum, enumCaseInsensitive
from cfnlint.rules import CloudFormationLintRule


class Enum(CloudFormationLintRule):
    """Check if parameters have a valid value"""

    id = "W2030"
    shortdesc = "Check if parameters have a valid value"
    description = (
        "Check if parameters have a valid value in case of an enumator. The Parameter's"
        " allowed values is based on the usages in property (Ref)"
    )
    source_url = "https://github.com/aws-cloudformation/cfn-lint/blob/main/docs/cfn-schema-specification.md#enum"
    tags = ["parameters", "resources", "property", "allowed value"]

    # This rule is triggered from the equivalent rule E3030
    # the values are fed from there and we adjust the error outputs
    # appropriately

    def enum(self, validator, enums, instance, schema):
        for err in enum(validator, enums, instance, schema):
            err.rule = self
            err.path_override = validator.context.path.value_path
            yield err

    def enumCaseInsensitive(self, validator, enums, instance, schema):
        for err in enumCaseInsensitive(validator, enums, instance, schema):
            err.rule = self
            err.path_override = validator.context.path.value_path
            yield err



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from collections import deque

import pytest

from cfnlint.context import Path, create_context_for_template
from cfnlint.helpers import FUNCTIONS
from cfnlint.jsonschema import CfnTemplateValidator
from cfnlint.template import Template


@pytest.fixture
def template(request):
    if hasattr(request, "param"):
        return request.param
    return {}


@pytest.fixture
def regions():
    return ["us-east-1"]


@pytest.fixture
def cfn(template, regions):
    return Template(
        "",
        template,
        regions,
    )


@pytest.fixture
def path(request):
    if hasattr(request, "param"):
        t_path = deque(request.param.get("path", []))
        value_path = deque(request.param.get("value_path", []))
        cfn_path = deque(request.param.get("cfn_path", []))

        return Path(
            path=t_path,
            value_path=value_path,
            cfn_path=cfn_path,
        )

    return Path(
        path=deque(),
        value_path=deque(),
        cfn_path=deque(),
    )


@pytest.fixture
def functions():
    return FUNCTIONS


@pytest.fixture
def strict_types(strict_types=True):
    return strict_types


@pytest.fixture
def context(cfn, path, functions, strict_types):
    return create_context_for_template(cfn).evolve(
        path=path,
        functions=functions,
        strict_types=strict_types,
    )


@pytest.fixture
def validators(request):
    if hasattr(request, "param"):
        return request.param
    return {}


@pytest.fixture
def validator(cfn, context, validators):
    return CfnTemplateValidator({}).extend(validators=validators)(
        context=context,
        cfn=cfn,
        schema={},
    )
